# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Methods to calculate various correlations
"""
from typing import Union
import math
import warnings
import numbers

from statistics import mean
import itertools as it
import numpy as np
import pandas as pd

from scipy.stats import pearsonr, spearmanr, kendalltau, t, norm
from scipy.optimize import brenth
from scipy.integrate import quad
from scipy import stats
from scipy.special import gamma, betaln, hyp2f1


def corr(x, y, tail='two-sided', method='pearson', **kwargs):
    """(Robust) correlation between two variables.
    This method was the patrt of pingouin package and was moved from it to METcalcpy

    Parameters
    ----------
    x, y : array_like
        First and second set of observations. ``x`` and ``y`` must be
        independent.
    tail : string
        Specify whether to return ``'one-sided'`` or ``'two-sided'`` p-value.
        Note that the former are simply half the latter.
    method : string
        Correlation type:

        * ``'pearson'``: Pearson :math:`r` product-moment correlation
        * ``'spearman'``: Spearman :math:`\\rho` rank-order correlation
        * ``'kendall'``: Kendall's :math:`\\tau_B` correlation
          (for ordinal data)
        * ``'bicor'``: Biweight midcorrelation (robust)
        * ``'percbend'``: Percentage bend correlation (robust)
        * ``'shepherd'``: Shepherd's pi correlation (robust)
    **kwargs : optional
        Optional argument(s) passed to the lower-level functions.

    Returns
    -------
    stats : :py:class:`pandas.DataFrame`

        * ``'n'``: Sample size (after removal of missing values)
        * ``'outliers'``: number of outliers, only if a robust method was used
        * ``'r'``: Correlation coefficient
        * ``'CI95'``: 95% parametric confidence intervals around :math:`r`
        * ``'r2'``: R-squared (:math:`= r^2`)
        * ``'adj_r2'``: Adjusted R-squared
        * ``'p-val'``: tail of the test
        * ``'BF10'``: Bayes Factor of the alternative hypothesis
          (only for Pearson correlation)
        * ``'power'``: achieved power of the test (= 1 - type II error).

    See also
    --------
    pairwise_corr : Pairwise correlation between columns of a pandas DataFrame
    partial_corr : Partial correlation
    rm_corr : Repeated measures correlation

    Notes
    -----
    The `Pearson correlation coefficient
    <https://en.wikipedia.org/wiki/Pearson_correlation_coefficient>`_
    measures the linear relationship between two datasets. Strictly speaking,
    Pearson's correlation requires that each dataset be normally distributed.
    Correlations of -1 or +1 imply a perfect negative and positive linear
    relationship, respectively, with 0 indicating the absence of association.

    .. math::
        r_{xy} = \\frac{\\sum_i(x_i - \\bar{x})(y_i - \\bar{y})}
        {\\sqrt{\\sum_i(x_i - \\bar{x})^2} \\sqrt{\\sum_i(y_i - \\bar{y})^2}}
        = \\frac{\\text{cov}(x, y)}{\\sigma_x \\sigma_y}

    where :math:`\\text{cov}` is the sample covariance and :math:`\\sigma`
    is the sample standard deviation.

    If ``method='pearson'``, The Bayes Factor is calculated using the
    :py:func:`pingouin.bayesfactor_pearson` function.

    The `Spearman correlation coefficient
    <https://en.wikipedia.org/wiki/Spearman%27s_rank_correlation_coefficient>`_
    is a non-parametric measure of the monotonicity of the relationship between
    two datasets. Unlike the Pearson correlation, the Spearman correlation does
    not assume that both datasets are normally distributed. Correlations of -1
    or +1 imply an exact negative and positive monotonic relationship,
    respectively. Mathematically, the Spearman correlation coefficient is
    defined as the Pearson correlation coefficient between the
    `rank variables <https://en.wikipedia.org/wiki/Ranking>`_.

    The `Kendall correlation coefficient
    <https://en.wikipedia.org/wiki/Kendall_rank_correlation_coefficient>`_
    is a measure of the correspondence between two rankings. Values also range
    from -1 (perfect disagreement) to 1 (perfect agreement), with 0 indicating
    the absence of association. Consistent with
    :py:func:`scipy.stats.kendalltau`, Pingouin returns the Tau-b coefficient,
    which adjusts for ties:

    .. math:: \\tau_B = \\frac{(P - Q)}{\\sqrt{(P + Q + T) (P + Q + U)}}

    where :math:`P` is the number of concordant pairs, :math:`Q` the number of
    discordand pairs, :math:`T` the number of ties in x, and :math:`U`
    the number of ties in y.

    The `biweight midcorrelation
    <https://en.wikipedia.org/wiki/Biweight_midcorrelation>`_ and
    percentage bend correlation [1]_ are both robust methods that
    protects against *univariate* outliers by down-weighting observations that
    deviate too much from the median.

    The Shepherd pi [2]_  correlation is
    both robust methods that returns the Spearman correlation coefficient after
    removing *bivariate* outliers. Briefly, the Shepherd pi uses a
    bootstrapping of the Mahalanobis distance to identify outliers.
     Note that this  method is
    significantly slower than the previous ones.

    .. important:: Please note that rows with missing values (NaN) are
        automatically removed.

    References
    ----------
    .. [1] Wilcox, R.R., 1994. The percentage bend correlation coefficient.
       Psychometrika 59, 601–616. https://doi.org/10.1007/BF02294395

    .. [2] Schwarzkopf, D.S., De Haas, B., Rees, G., 2012. Better ways to
       improve standards in brain-behavior correlation analysis. Front.
       Hum. Neurosci. 6, 200. https://doi.org/10.3389/fnhum.2012.00200

    .. [3] Rousselet, G.A., Pernet, C.R., 2012. Improving standards in
       brain-behavior correlation analyses. Front. Hum. Neurosci. 6, 119.
       https://doi.org/10.3389/fnhum.2012.00119

    .. [4] Pernet, C.R., Wilcox, R., Rousselet, G.A., 2012. Robust correlation
       analyses: false positive and power validation using a new open
       source matlab toolbox. Front. Psychol. 3, 606.
       https://doi.org/10.3389/fpsyg.2012.00606

    Examples
    --------
    1. Pearson correlation

    >>> import numpy as np
    >>> # Generate random correlated samples
    >>> np.random.seed(123)
    >>> mean, cov = [4, 6], [(1, .5), (.5, 1)]
    >>> x, y = np.random.multivariate_normal(mean, cov, 30).T
    >>> # Compute Pearson correlation
    >>> pg.corr(x, y).round(3)
              n      r         CI95%     r2  adj_r2  p-val  BF10  power
    pearson  30  0.491  [0.16, 0.72]  0.242   0.185  0.006  8.55  0.809

    2. Pearson correlation with two outliers

    >>> x[3], y[5] = 12, -8
    >>> pg.corr(x, y).round(3)
              n      r          CI95%     r2  adj_r2  p-val   BF10  power
    pearson  30  0.147  [-0.23, 0.48]  0.022  -0.051  0.439  0.302  0.121

    3. Spearman correlation (robust to outliers)

    >>> pg.corr(x, y, method="spearman").round(3)
               n      r         CI95%     r2  adj_r2  p-val  power
    spearman  30  0.401  [0.05, 0.67]  0.161   0.099  0.028   0.61

    4. Biweight midcorrelation (robust)

    >>> pg.corr(x, y, method="bicor").round(3)
            n      r         CI95%     r2  adj_r2  p-val  power
    bicor  30  0.393  [0.04, 0.66]  0.155   0.092  0.031  0.592

    5. Percentage bend correlation (robust)

    >>> pg.corr(x, y, method='percbend').round(3)
               n      r         CI95%     r2  adj_r2  p-val  power
    percbend  30  0.389  [0.03, 0.66]  0.151   0.089  0.034  0.581

    6. Shepherd's pi correlation (robust)

    >>> pg.corr(x, y, method='shepherd').round(3)
               n  outliers      r         CI95%     r2  adj_r2  p-val  power
    shepherd  30         2  0.437  [0.09, 0.69]  0.191   0.131   0.02  0.694


    8. One-tailed Pearson correlation

    >>> pg.corr(x, y, tail="one-sided", method='pearson').round(3)
              n      r          CI95%     r2  adj_r2  p-val   BF10  power
    pearson  30  0.147  [-0.23, 0.48]  0.022  -0.051   0.22  0.467  0.194

    9. Using columns of a pandas dataframe

    >>> import pandas as pd
    >>> data = pd.DataFrame({'x': x, 'y': y})
    >>> pg.corr(data['x'], data['y']).round(3)
              n      r          CI95%     r2  adj_r2  p-val   BF10  power
    pearson  30  0.147  [-0.23, 0.48]  0.022  -0.051  0.439  0.302  0.121
    """
    # Safety check
    x = np.asarray(x)
    y = np.asarray(y)
    assert x.ndim == y.ndim == 1, 'x and y must be 1D array.'
    assert x.size == y.size, 'x and y must have the same length.'
    _msg = 'tail must be "two-sided" or "one-sided".'
    assert tail in ['two-sided', 'one-sided'], _msg

    # Remove rows with missing values
    x, y = remove_na(x, y, paired=True)
    nx = x.size

    # Compute correlation coefficient
    if method == 'pearson':
        r, pval = pearsonr(x, y)
    elif method == 'spearman':
        r, pval = spearmanr(x, y, **kwargs)
    elif method == 'kendall':
        r, pval = kendalltau(x, y, **kwargs)
    elif method == 'bicor':
        r, pval = bicor(x, y, **kwargs)
    elif method == 'percbend':
        r, pval = percbend(x, y, **kwargs)
    elif method == 'shepherd':
        r, pval, outliers = shepherd(x, y, **kwargs)
    else:
        raise ValueError(f'Method "{method}" not recognized.')

    if np.isnan(r):
        # Correlation failed -- new in version v0.3.4, instead of raising an
        # error we just return a dataframe full of NaN (except sample size).
        # This avoid sudden stop in pingouin.pairwise_corr.
        return pd.DataFrame({'n': nx, 'r': np.nan, 'CI95%': np.nan,
                             'r2': np.nan, 'adj_r2': np.nan, 'p-val': np.nan,
                             'BF10': np.nan, 'power': np.nan}, index=[method])

    # Compute r2 and adj_r2
    r2 = r ** 2
    adj_r2 = 1 - (((1 - r2) * (nx - 1)) / (nx - 3))

    # Compute the parametric 95% confidence interval and power
    ci = compute_esci(stat=r, nx=nx, ny=nx, eftype='r', decimals=6)
    pr = power_corr(r=r, n=nx, power=None, alpha=0.05, tail=tail)

    # Create dictionary
    stats_dict = \
        {'n': nx,
         'r': r,
         'r2': r2,
         'adj_r2': adj_r2,
         'CI95%': [ci],
         'p-val': pval if tail == 'two-sided' else .5 * pval,
         'power': pr
         }

    # Compute the BF10 for Pearson correlation only
    if method == 'pearson':
        stats_dict['BF10'] = bayesfactor_pearson(r, nx, tail=tail)

    # Convert to DataFrame
    stats_df = pd.DataFrame.from_records(stats_dict, index=[method])

    # Define order
    col_keep = ['n', 'outliers', 'r', 'CI95%', 'r2', 'adj_r2', 'p-val',
                'BF10', 'power']
    col_order = [k for k in col_keep if k in stats_df.keys().tolist()]
    return _postprocess_dataframe(stats_df)[col_order]


def acf(x: Union[list, np.array], acf_type: str = 'correlation', lag_max: Union[int, None] = None) \
        -> Union[list, None]:
    """
    The function acf computes estimates of the autocovariance or autocorrelation function.
    Similar to Rscript function

    Args:
        x - numeric array.
        acf_type - character string giving the type of acf to be computed.
                Allowed values are "correlation" (the default), "covariance".
        lag_max - maximum lag at which to calculate the acf.
                Can be an integer or None (the default).
                Default is 10*log10(N/m) where N is the number of
                observations and m the number of series.
                Will be automatically limited to one less than the number of
                observations in the series.

        :return: - An array with the same dimensions as lag_max containing the estimated acf
    """
    # validate acf type
    if acf_type not in ['covariance', 'correlation']:
        print('ERROR  incorrect acf_type')
        return None
    if x is None or len(x) == 0:
        return None

    acf_result = []
    size = np.size(x)

    # calculate mean of the array excluding None
    mean_val = mean(remove_none(x))

    # calculate lag if not provided
    if lag_max is None:
        lag_max = math.floor(10 * (math.log10(size) - math.log10(1)))
    lag_max = int(min(lag_max, size - 1))

    cov_0 = autocovariance(x, size, 0, mean_val)
    for i in range(lag_max+1):
        cov = autocovariance(x, size, i, mean_val)
        if acf_type == 'covariance':
            acf_result.append(cov)
        elif acf_type == 'correlation':
            acf_result.append(cov / cov_0)
    return acf_result


def autocovariance(x, list_size, n_lag, mean_val):
    """
    The function that computes  autocovariance for the first n_lag elements of the list
    :param x: numeric array
    :param list_size: size of list
    :param n_lag: lag at which to calculate the autocovariance
    :param mean_val:  mean value of the array
    :return:  autocovariance
    """
    total_autocov = 0
    count_autocov = 0
    for i in np.arange(0, list_size - n_lag):
        if x[i + n_lag] is not None and x[i] is not None and not math.isnan(x[i + n_lag]) and not math.isnan(x[i]):
            total_autocov += ((x[i + n_lag]) - mean_val) * (x[i] - mean_val)
            count_autocov = count_autocov + 1
    return (1 / (count_autocov + n_lag)) * total_autocov


def bicor(x, y, c=9):
    """
    Biweight midcorrelation.

    Parameters
    ----------
    x, y : array_like
        First and second set of observations. x and y must be independent.
    c : float
        Tuning constant for the biweight estimator (default = 9.0).

    Returns
    -------
    r : float
        Correlation coefficient.
    pval : float
        Two-tailed p-value.

    Notes
    -----
    This function will return (np.nan, np.nan) if mad(x) == 0 or mad(y) == 0.

    References
    ----------
    https://en.wikipedia.org/wiki/Biweight_midcorrelation

    https://docs.astropy.org/en/stable/api/astropy.stats.biweight.biweight_midcovariance.html

    Langfelder, P., & Horvath, S. (2012). Fast R Functions for Robust
    Correlations and Hierarchical Clustering. Journal of Statistical Software,
    46(11). https://www.ncbi.nlm.nih.gov/pubmed/23050260
    """

    # Calculate median
    nx = x.size
    x_median = np.median(x)
    y_median = np.median(y)
    # Raw median absolute deviation
    x_mad = np.median(np.abs(x - x_median))
    y_mad = np.median(np.abs(y - y_median))
    if x_mad == 0 or y_mad == 0:
        # From Langfelder and Horvath 2012:
        # "Strictly speaking, a call to bicor in R should return a missing
        # value if mad(x) = 0 or mad(y) = 0." This avoids division by zero.
        return np.nan, np.nan
    # Calculate weights
    u = (x - x_median) / (c * x_mad)
    v = (y - y_median) / (c * y_mad)
    w_x = (1 - u ** 2) ** 2 * ((1 - np.abs(u)) > 0)
    w_y = (1 - v ** 2) ** 2 * ((1 - np.abs(v)) > 0)
    # Normalize x and y by weights
    x_norm = (x - x_median) * w_x
    y_norm = (y - y_median) * w_y
    denom = (np.sqrt((x_norm ** 2).sum()) * np.sqrt((y_norm ** 2).sum()))
    # Calculate r, t and two-sided p-value
    r = (x_norm * y_norm).sum() / denom
    tval = r * np.sqrt((nx - 2) / (1 - r ** 2))
    pval = 2 * t.sf(abs(tval), nx - 2)
    return r, pval


def percbend(x, y, beta=.2):
    """
    Percentage bend correlation (Wilcox 1994).

    Parameters
    ----------
    x, y : array_like
        First and second set of observations. x and y must be independent.
    beta : float
        Bending constant for omega (0 <= beta <= 0.5).

    Returns
    -------
    r : float
        Percentage bend correlation coefficient.
    pval : float
        Two-tailed p-value.

    Notes
    -----
    Code inspired by Matlab code from Cyril Pernet and Guillaume Rousselet.

    References
    ----------
    .. [1] Wilcox, R.R., 1994. The percentage bend correlation coefficient.
       Psychometrika 59, 601–616. https://doi.org/10.1007/BF02294395

    .. [2] Pernet CR, Wilcox R, Rousselet GA. Robust Correlation Analyses:
       False Positive and Power Validation Using a New Open Source Matlab
       Toolbox. Frontiers in Psychology. 2012;3:606.
       doi:10.3389/fpsyg.2012.00606.
    """
    X = np.column_stack((x, y))
    nx = X.shape[0]
    M = np.tile(np.median(X, axis=0), nx).reshape(X.shape)
    W = np.sort(np.abs(X - M), axis=0)
    m = int((1 - beta) * nx)
    omega = W[m - 1, :]
    P = (X - M) / omega
    P[np.isinf(P)] = 0
    P[np.isnan(P)] = 0

    # Loop over columns
    a = np.zeros((2, nx))
    for c in [0, 1]:
        psi = P[:, c]
        i1 = np.where(psi < -1)[0].size
        i2 = np.where(psi > 1)[0].size
        s = X[:, c].copy()
        s[np.where(psi < -1)[0]] = 0
        s[np.where(psi > 1)[0]] = 0
        pbos = (np.sum(s) + omega[c] * (i2 - i1)) / (s.size - i1 - i2)
        a[c] = (X[:, c] - pbos) / omega[c]

    # Bend
    a[a <= -1] = -1
    a[a >= 1] = 1

    # Get r, tval and pval
    a, b = a
    r = (a * b).sum() / np.sqrt((a ** 2).sum() * (b ** 2).sum())
    tval = r * np.sqrt((nx - 2) / (1 - r ** 2))
    pval = 2 * t.sf(abs(tval), nx - 2)
    return r, pval


def shepherd(x, y, n_boot=200):
    """
    Shepherd's Pi correlation, equivalent to Spearman's rho after outliers
    removal.

    Parameters
    ----------
    x, y : array_like
        First and second set of observations. x and y must be independent.
    n_boot : int
        Number of bootstrap samples to calculate.

    Returns
    -------
    r : float
        Pi correlation coefficient
    pval : float
        Two-tailed adjusted p-value.
    outliers : array of bool
        Indicate if value is an outlier or not

    Notes
    -----
    It first bootstraps the Mahalanobis distances, removes all observations
    with m >= 6 and finally calculates the correlation of the remaining data.

    Pi is Spearman's Rho after outlier removal.
    """
    X = np.column_stack((x, y))
    # Bootstrapping on Mahalanobis distance
    m = bsmahal(X, X, n_boot)
    # Determine outliers
    outliers = (m >= 6)
    # Compute correlation
    r, pval = spearmanr(x[~outliers], y[~outliers])
    # (optional) double the p-value to achieve a nominal false alarm rate
    # pval *= 2
    # pval = 1 if pval > 1 else pval
    return r, pval, outliers


def bsmahal(a, b, n_boot=200):
    """
    Bootstraps Mahalanobis distances for Shepherd's pi correlation.

    Parameters
    ----------
    a : ndarray (shape=(n, 2))
        Data
    b : ndarray (shape=(n, 2))
        Data
    n_boot : int
        Number of bootstrap samples to calculate.

    Returns
    -------
    m : ndarray (shape=(n,))
        Mahalanobis distance for each row in a, averaged across all the
        bootstrap resamples.
    """
    n, m = b.shape
    MD = np.zeros((n, n_boot))
    nr = np.arange(n)
    xB = np.random.choice(nr, size=(n_boot, n), replace=True)
    # Bootstrap the MD
    for i in np.arange(n_boot):
        s1 = b[xB[i, :], 0]
        s2 = b[xB[i, :], 1]
        X = np.column_stack((s1, s2))
        mu = X.mean(0)
        _, R = np.linalg.qr(X - mu)
        sol = np.linalg.solve(R.T, (a - mu).T)
        MD[:, i] = np.sum(sol ** 2, 0) * (n - 1)
    # Average across all bootstraps
    return MD.mean(1)


def bayesfactor_pearson(r, n, tail='two-sided', method='ly', kappa=1.):
    """
    Bayes Factor of a Pearson correlation.

    Parameters
    ----------
    r : float
        Pearson correlation coefficient.
    n : int
        Sample size.
    tail : float
        Tail of the alternative hypothesis. Can be *'two-sided'*
    method : str
        Method to compute the Bayes Factor. Can be *'ly'* (default) or
        *'wetzels'*. The former has an exact analytical solution, while the
        latter requires integral solving (and is therefore slower). *'wetzels'*
        was the default in Pingouin <= 0.2.5. See Notes for details.
    kappa : float
        Kappa factor. This is sometimes called the *rscale* parameter, and
        is only used when ``method`` is *'ly'*.

    Returns
    -------
    bf : float
        Bayes Factor (BF10).
        The Bayes Factor quantifies the evidence in favour of the alternative
        hypothesis.

    See also
    --------
    corr : (Robust) correlation between two variables
    pairwise_corr : Pairwise correlation between columns of a pandas DataFrame
    bayesfactor_ttest : Bayes Factor of a T-test
    bayesfactor_binom : Bayes Factor of a binomial test

    Notes
    -----
    To compute the Bayes Factor directly from the raw data, use the
    :py:func:`pingouin.corr` function.

    The two-sided **Wetzels Bayes Factor** (also called *JZS Bayes Factor*)
    is calculated using the equation 13 and associated R code of [1]_:

    .. math::

        \\text{BF}_{10}(n, r) = \\frac{\\sqrt{n/2}}{\\gamma(1/2)}*
        \\int_{0}^{\\infty}e((n-2)/2)*
        log(1+g)+(-(n-1)/2)log(1+(1-r^2)*g)+(-3/2)log(g)-n/2g

    where :math:`n` is the sample size, :math:`r` is the Pearson correlation
    coefficient and :math:`g` is is an auxiliary variable that is integrated
    out numerically. Since the Wetzels Bayes Factor requires solving an
    integral, it is slower than the analytical solution described below.

    The two-sided **Ly Bayes Factor** (also called *Jeffreys
    exact Bayes Factor*) is calculated using equation 25 of [2]_:

    .. math::

        \\text{BF}_{10;k}(n, r) = \\frac{2^{\\frac{k-2}{k}}\\sqrt{\\pi}}
        {\\beta(\\frac{1}{k}, \\frac{1}{k})} \\cdot
        \\frac{\\Gamma(\\frac{2+k(n-1)}{2k})}{\\Gamma(\\frac{2+nk}{2k})}
        \\cdot 2F_1(\\frac{n-1}{2}, \\frac{n-1}{2}, \\frac{2+nk}{2k}, r^2)

    The one-sided version is described in eq. 27 and 28 of Ly et al, 2016.
    Please take note that the one-sided test requires the
    `mpmath <http://mpmath.org/>`_ package.

    Results have been validated against JASP and the BayesFactor R package.

    References
    ----------
    .. [1] Ly, A., Verhagen, J. & Wagenmakers, E.-J. Harold Jeffreys’s default
       Bayes factor hypothesis tests: Explanation, extension, and
       application in psychology. J. Math. Psychol. 72, 19–32 (2016).

    .. [2] Wetzels, R. & Wagenmakers, E.-J. A default Bayesian hypothesis test
       for correlations and partial correlations. Psychon. Bull. Rev. 19,
       1057–1064 (2012).

    Examples
    --------
    Bayes Factor of a Pearson correlation

    >>> r, n = 0.6, 20
    >>> bf = bayesfactor_pearson(r, n)
    >>> print("Bayes Factor: %.3f" % bf)
    Bayes Factor: 10.634

    Compare to Wetzels method:

    >>> bf = bayesfactor_pearson(r, n, method='wetzels')
    >>> print("Bayes Factor: %.3f" % bf)
    Bayes Factor: 8.221


    """
    assert method.lower() in ['ly', 'wetzels'], 'Method not recognized.'
    assert tail.lower() in ['two-sided', 'one-sided', 'greater', 'less',
                            'g', 'l', 'positive', 'negative', 'pos', 'neg']

    # Wrong input
    if not np.isfinite(r) or n < 2:
        return np.nan
    assert -1 <= r <= 1, 'r must be between -1 and 1.'

    if tail.lower() != 'two-sided' and method.lower() == 'wetzels':
        warnings.warn("One-sided Bayes Factor are not supported by the "
                      "Wetzels's method. Switching to method='ly'.")
        method = 'ly'

    if method.lower() == 'wetzels':
        # Wetzels & Wagenmakers, 2012. Integral solving

        def fun(g, r, n):
            return math.exp(((n - 2) / 2) * math.log(1 + g) + (-(n - 1) / 2)
                            * math.log(1 + (1 - r ** 2) * g) + (-3 / 2)
                            * math.log(g) + - n / (2 * g))

        integr = quad(fun, 0, np.inf, args=(r, n))[0]
        bf10 = np.sqrt((n / 2)) / gamma(1 / 2) * integr

    else:
        # Ly et al, 2016. Analytical solution.
        k = kappa
        lbeta = betaln(1 / k, 1 / k)
        log_hyperterm = math.log(hyp2f1(((n - 1) / 2), ((n - 1) / 2),
                                        ((n + 2 / k) / 2), r ** 2))
        bf10 = math.exp((1 - 2 / k) * math.log(2) + 0.5 * math.log(math.pi) - lbeta
                        + math.lgamma((n + 2 / k - 1) / 2) - math.lgamma((n + 2 / k) / 2) +
                        log_hyperterm)

    return bf10


def compute_esci(stat=None, nx=None, ny=None, paired=False, eftype='cohen',
                 confidence=.95, decimals=2):
    """Parametric confidence intervals around a Cohen d or a
    correlation coefficient.

    Parameters
    ----------
    stat : float
        Original effect size. Must be either a correlation coefficient or a
        Cohen-type effect size (Cohen d or Hedges g).
    nx, ny : int
        Length of vector x and y.
    paired : bool
        Indicates if the effect size was estimated from a paired sample.
        This is only relevant for cohen or hedges effect size.
    eftype : string
        Effect size type. Must be ``'r'`` (correlation) or ``'cohen'``
        (Cohen d or Hedges g).
    confidence : float
        Confidence level (0.95 = 95%)
    decimals : int
        Number of rounded decimals.

    Returns
    -------
    ci : array
        Desired converted effect size

    Notes
    -----
    To compute the parametric confidence interval around a
    **Pearson r correlation** coefficient, one must first apply a
    Fisher's r-to-z transformation:

    .. math:: z = 0.5 \\cdot \\ln \\frac{1 + r}{1 - r} = \\text{arctanh}(r)

    and compute the standard error:

    .. math:: \\text{SE} = \\frac{1}{\\sqrt{n - 3}}

    where :math:`n` is the sample size.

    The lower and upper confidence intervals - *in z-space* - are then
    given by:

    .. math:: \\text{ci}_z = z \\pm \\text{crit} \\cdot \\text{SE}

    where :math:`\\text{crit}` is the critical value of the normal distribution
    corresponding to the desired confidence level (e.g. 1.96 in case of a 95%
    confidence interval).

    These confidence intervals can then be easily converted back to *r-space*:

    .. math::

        \\text{ci}_r = \\frac{\\exp(2 \\cdot \\text{ci}_z) - 1}
        {\\exp(2 \\cdot \\text{ci}_z) + 1} = \\text{tanh}(\\text{ci}_z)

    A formula for calculating the confidence interval for a
    **Cohen d effect size** is given by Hedges and Olkin (1985, p86).
    If the effect size estimate from the sample is :math:`d`, then it follows a
    T distribution with standard error:

    .. math::

        \\text{SE} = \\sqrt{\\frac{n_x + n_y}{n_x \\cdot n_y} +
        \\frac{d^2}{2 (n_x + n_y)}}

    where :math:`n_x` and :math:`n_y` are the sample sizes of the two groups.

    In one-sample test or paired test, this becomes:

    .. math::

        \\text{SE} = \\sqrt{\\frac{1}{n_x} + \\frac{d^2}{2 n_x}}

    The lower and upper confidence intervals are then given by:

    .. math:: \\text{ci}_d = d \\pm \\text{crit} \\cdot \\text{SE}

    where :math:`\\text{crit}` is the critical value of the T distribution
    corresponding to the desired confidence level.

    References
    ----------
    * https://en.wikipedia.org/wiki/Fisher_transformation

    * Hedges, L., and Ingram Olkin. "Statistical models for meta-analysis."
      (1985).

    * http://www.leeds.ac.uk/educol/documents/00002182.htm

    * https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5133225/

    Examples
    --------
    1. Confidence interval of a Pearson correlation coefficient

    >>> x = [3, 4, 6, 7, 5, 6, 7, 3, 5, 4, 2]
    >>> y = [4, 6, 6, 7, 6, 5, 5, 2, 3, 4, 1]
    >>> nx, ny = len(x), len(y)
    >>> stat = pg.compute_effsize(x, y, eftype='r')
    >>> ci = pg.compute_esci(stat=stat, nx=nx, ny=ny, eftype='r')
    >>> print(round(stat, 4), ci)
    0.7468 [0.27 0.93]

    2. Confidence interval of a Cohen d

    >>> stat = pg.compute_effsize(x, y, eftype='cohen')
    >>> ci = pg.compute_esci(stat, nx=nx, ny=ny, eftype='cohen', decimals=3)
    >>> print(round(stat, 4), ci)
    0.1538 [-0.737  1.045]
    """
    assert eftype.lower() in ['r', 'pearson', 'spearman', 'cohen',
                              'd', 'g', 'hedges']
    assert stat is not None and nx is not None
    assert isinstance(confidence, float)
    assert 0 < confidence < 1, 'confidence must be between 0 and 1.'

    if eftype.lower() in ['r', 'pearson', 'spearman']:
        z = np.arctanh(stat)  # R-to-z transform
        se = 1 / np.sqrt(nx - 3)
        crit = np.abs(norm.ppf((1 - confidence) / 2))
        ci_z = np.array([z - crit * se, z + crit * se])
        ci = np.tanh(ci_z)  # Transform back to r
    else:
        # Cohen d. Results are different than JASP which uses a non-central T
        # distribution. See github.com/jasp-stats/jasp-issues/issues/525
        if ny == 1 or paired:
            # One-sample or paired. Results vary slightly from the cohen.d R
            # function which uses instead:
            # >>> sqrt((n / (n / 2)^2) + .5*(dd^2 / n)) -- one-sample
            # >>> sqrt( (1/n1 + dd^2/(2*n1))*(2-2*r)); -- paired
            # where r is the correlation between samples
            # https://github.com/mtorchiano/effsize/blob/master/R/CohenD.R
            # However, Pingouin uses the formulas on www.real-statistics.com
            se = np.sqrt(1 / nx + stat ** 2 / (2 * nx))
            dof = nx - 1
        else:
            # Independent two-samples: give same results as R:
            # >>> cohen.d(..., paired = FALSE, noncentral=FALSE)
            se = np.sqrt(((nx + ny) / (nx * ny)) + (stat ** 2) / (2 * (nx + ny)))
            dof = nx + ny - 2
        crit = np.abs(t.ppf((1 - confidence) / 2, dof))
        ci = np.array([stat - crit * se, stat + crit * se])
    return np.round(ci, decimals)


def power_corr(r=None, n=None, power=None, alpha=0.05, tail='two-sided'):
    """
    Evaluate power, sample size, correlation coefficient or
    significance level of a correlation test.

    Parameters
    ----------
    r : float
        Correlation coefficient.
    n : int
        Number of observations (sample size).
    power : float
        Test power (= 1 - type II error).
    alpha : float
        Significance level (type I error probability).
        The default is 0.05.
    tail : str
        Indicates whether the test is `'two-sided'` or `'one-sided'`.

    Notes
    -----
    Exactly ONE of the parameters ``r``, ``n``, ``power`` and ``alpha`` must
    be passed as None, and that parameter is determined from the others.

    Notice that ``alpha`` has a default value of 0.05 so None must be
    explicitly passed if you want to compute it.

    :py:func:`scipy.optimize.brenth` is used to solve power equations for other
    variables (i.e. sample size, effect size, or significance level). If the
    solving fails, a nan value is returned.

    This function is a Python adaptation of the `pwr.r.test`
    function implemented in the
    `pwr <https://cran.r-project.org/web/packages/pwr/pwr.pdf>`_ R package.

    Examples
    --------
    1. Compute achieved power given ``r``, ``n`` and ``alpha``

    >>> print('power: %.4f' % power_corr(r=0.5, n=20))
    power: 0.6379

    2. Compute required sample size given ``r``, ``power`` and ``alpha``

    >>> print('n: %.4f' % power_corr(r=0.5, power=0.80,
    ...                                tail='one-sided'))
    n: 22.6091

    3. Compute achieved ``r`` given ``n``, ``power`` and ``alpha`` level

    >>> print('r: %.4f' % power_corr(n=20, power=0.80, alpha=0.05))
    r: 0.5822

    4. Compute achieved alpha level given ``r``, ``n`` and ``power``

    >>> print('alpha: %.4f' % power_corr(r=0.5, n=20, power=0.80,
    ...                                    alpha=None))
    alpha: 0.1377
    """
    # Check the number of arguments that are None
    n_none = sum([v is None for v in [r, n, power, alpha]])
    if n_none != 1:
        raise ValueError('Exactly one of n, r, power, and alpha must be None')

    # Safety checks
    if r is not None:
        assert -1 <= r <= 1
        r = abs(r)
    if alpha is not None:
        assert 0 < alpha <= 1
    if power is not None:
        assert 0 < power <= 1
    if n is not None:
        if n <= 4:
            warnings.warn("Sample size is too small to estimate power "
                          "(n <= 4). Returning NaN.")
            return np.nan

    # Define main function
    if tail == 'two-sided':

        def func(r, n, power, alpha):
            dof = n - 2
            ttt = stats.t.ppf(1 - alpha / 2, dof)
            rc = np.sqrt(ttt ** 2 / (ttt ** 2 + dof))
            zr = np.arctanh(r) + r / (2 * (n - 1))
            zrc = np.arctanh(rc)
            power = stats.norm.cdf((zr - zrc) * np.sqrt(n - 3)) + \
                    stats.norm.cdf((-zr - zrc) * np.sqrt(n - 3))
            return power

    else:

        def func(r, n, power, alpha):
            dof = n - 2
            ttt = stats.t.ppf(1 - alpha, dof)
            rc = np.sqrt(ttt ** 2 / (ttt ** 2 + dof))
            zr = np.arctanh(r) + r / (2 * (n - 1))
            zrc = np.arctanh(rc)
            power = stats.norm.cdf((zr - zrc) * np.sqrt(n - 3))
            return power

    # Evaluate missing variable
    if power is None and n is not None and r is not None:
        # Compute achieved power given r, n and alpha
        return func(r, n, power=None, alpha=alpha)

    elif n is None and power is not None and r is not None:
        # Compute required sample size given r, power and alpha

        def _eval_n(n, r, power, alpha):
            return func(r, n, power, alpha) - power

        try:
            return brenth(_eval_n, 4 + 1e-10, 1e+09, args=(r, power, alpha))
        except ValueError:  # pragma: no cover
            return np.nan

    elif r is None and power is not None and n is not None:
        # Compute achieved r given sample size, power and alpha level

        def _eval_r(r, n, power, alpha):
            return func(r, n, power, alpha) - power

        try:
            return brenth(_eval_r, 1e-10, 1 - 1e-10, args=(n, power, alpha))
        except ValueError:  # pragma: no cover
            return np.nan

    else:
        # Compute achieved alpha (significance) level given r, n and power

        def _eval_alpha(alpha, r, n, power):
            return func(r, n, power, alpha) - power

        try:
            return brenth(_eval_alpha, 1e-10, 1 - 1e-10, args=(r, n, power))
        except ValueError:  # pragma: no cover
            return np.nan


def _postprocess_dataframe(df):
    """Apply some post-processing to an ouput dataframe (e.g. rounding).

    Whether and how rounding is applied is governed by options specified in
    `pingouin.options`. The default rounding (number of decimals) is
    determined by `pingouin.options['round']`. You can specify rounding for a
    given column name by the option `'round.column.<colname>'`, e.g.
    `'round.column.CI95%'`. Analogously, `'round.row.<rowname>'` also works
    (where `rowname`) refers to the pandas index), as well as
    `'round.cell.[<rolname>]x[<colname]'`. A cell-based option is used,
    if available; if not, a column-based option is used, if
    available; if not, a row-based option is used, if available; if not,
    the default is used. (Default `pingouin.options['round'] = None`,
    i.e. no rounding is applied.)

    If a round option is `callable` instead of `int`, then it will be called,
    and the return value stored in the cell.

    Post-processing is applied on a copy of the DataFrame, leaving the
    original DataFrame untouched.

    This is an internal function (no public API).

    Parameters
    ----------
    df : :py:class:`pandas.DataFrame`
        Dataframe to apply post-processing to (e.g. ANOVA summary)

    Returns
    ----------
    df : :py:class:`pandas.DataFrame`
        Dataframe with post-processing applied
    """
    df = df.copy()
    for row, col in it.product(df.index, df.columns):
        round_option = _get_round_setting_for(row, col)
        if round_option is None:
            continue
        if callable(round_option):
            newval = round_option(df.at[row, col])
            # ensure that dtype changes are processed
            df[col] = df[col].astype(type(newval))
            df.at[row, col] = newval
            continue
        if isinstance(df.at[row, col], bool):
            # No rounding if value is a boolean
            continue
        is_number = isinstance(df.at[row, col], numbers.Number)
        is_array = isinstance(df.at[row, col], np.ndarray)
        if not any([is_number, is_array]):
            # No rounding if value is not a Number or an array
            continue
        if is_array:
            is_float_array = issubclass(df.at[row, col].dtype.type,
                                        np.floating)
            if not is_float_array:
                # No rounding if value is not a float array
                continue
        df.at[row, col] = np.round(df.at[row, col], decimals=round_option)
    return df


def _format_bf(bf, precision=3, trim='0'):
    """Format BF10 to floating point or scientific notation.
    """
    if type(bf) == str:
        return bf
    if bf >= 1e4 or bf <= 1e-4:
        out = np.format_float_scientific(bf, precision=precision, trim=trim)
    else:
        out = np.format_float_positional(bf, precision=precision, trim=trim)
    return out


def _get_round_setting_for(row, col):
    options = {
        'round': None,
        'round.column.CI95%': 2,
        'round.column.BF10': _format_bf
    }
    keys_to_check = (
        'round.cell.[{}]x[{}]'.format(row, col),
        'round.column.{}'.format(col), 'round.row.{}'.format(row))
    for key in keys_to_check:
        try:
            return options[key]
        except KeyError:
            pass
    return options['round']


def remove_na(x, y=None, paired=False, axis='rows'):
    """Remove missing values along a given axis in one or more (paired) numpy
    arrays.

    Parameters
    ----------
    x, y : 1D or 2D np.ndarray arrays
        Data. ``x`` and ``y`` must have the same number of dimensions.
        ``y`` can be None to only remove missing values in ``x``.
    paired : bool
        Indicates if the measurements are paired or not.
    axis : str
        Axis or axes along which missing values are removed.
        Can be 'rows' or 'columns'. This has no effect if ``x`` and ``y`` are
        one-dimensional arrays.

    Returns
    -------
    x, y : np.ndarray
        Data without missing values

    Examples
    --------
    Single 1D array

    >>> import numpy as np
    >>> x = [6.4, 3.2, 4.5, np.nan]
    >>> remove_na(x)
    array([6.4, 3.2, 4.5])

    With two paired 1D arrays

    >>> y = [2.3, np.nan, 5.2, 4.6]
    >>> remove_na(x, y, paired=True)
    (array([6.4, 4.5]), array([2.3, 5.2]))

    With two independent 2D arrays

    >>> x = np.array([[4, 2], [4, np.nan], [7, 6]])
    >>> y = np.array([[6, np.nan], [3, 2], [2, 2]])
    >>> x_no_nan, y_no_nan = remove_na(x, y, paired=False)
    """
    # Safety checks
    x = np.asarray(x)
    assert x.size > 1, 'x must have more than one element.'
    assert axis in ['rows', 'columns'], 'axis must be rows or columns.'

    if y is None:
        return _remove_na_single(x, axis=axis)
    elif isinstance(y, (int, float, str)):
        return _remove_na_single(x, axis=axis), y
    else:  # y is list, np.array, pd.Series
        y = np.asarray(y)
        # Make sure that we just pass-through if y have only 1 element
        if y.size == 1:
            return _remove_na_single(x, axis=axis), y
        if x.ndim != y.ndim or paired is False:
            # x and y do not have the same dimension
            x_no_nan = _remove_na_single(x, axis=axis)
            y_no_nan = _remove_na_single(y, axis=axis)
            return x_no_nan, y_no_nan

    # At this point, we assume that x and y are paired and have same dimensions
    if x.ndim == 1:
        # 1D arrays
        x_mask = ~np.isnan(x)
        y_mask = ~np.isnan(y)
    else:
        # 2D arrays
        ax = 1 if axis == 'rows' else 0
        x_mask = ~np.any(np.isnan(x), axis=ax)
        y_mask = ~np.any(np.isnan(y), axis=ax)

    # Check if missing values are present
    if ~x_mask.all() or ~y_mask.all():
        ax = 0 if axis == 'rows' else 1
        ax = 0 if x.ndim == 1 else ax
        both = np.logical_and(x_mask, y_mask)
        x = x.compress(both, axis=ax)
        y = y.compress(both, axis=ax)
    return x, y


def _remove_na_single(x, axis='rows'):
    """Remove NaN in a single np.ndarray numpy array.
    This is an internal Pingouin function.
    """
    if x.ndim == 1:
        # 1D arrays
        x_mask = ~np.isnan(x)
    else:
        # 2D arrays
        ax = 1 if axis == 'rows' else 0
        x_mask = ~np.any(np.isnan(x), axis=ax)
    # Check if missing values are present
    if ~x_mask.all():
        ax = 0 if axis == 'rows' else 1
        ax = 0 if x.ndim == 1 else ax
        x = x.compress(x_mask, axis=ax)
    return x


def remove_none(x):
    """ Remove missing (None, nan) values from the list
    :param x: numeric list
    :return: a list without None  or empty array
    """
    if x is None:
        return []

    return [elem for elem in x if elem is not None and not pd.isna(elem)]



import argparse
import math
import numpy as np

if __name__ == '__main__':

    nx_default = 10
    ny_default = 10

    parser = argparse.ArgumentParser()
    parser.add_argument('--nx', type=int,
                        default=nx_default,
                        help='Number of x grid points')
    parser.add_argument('--ny', type=int,
                        default=ny_default,
                        help='Number of y grid points')
    parser.add_argument('--x_min', type=float,
                        default=0,
                        help='Domain left x-coordinate')
    parser.add_argument('--x_max', type=float,
                        default=nx_default,
                        help='Domain right x-coordinate')
    parser.add_argument('--y_min', type=float,
                        default=0,
                        help='Domain bottom y-coordinate')
    parser.add_argument('--y_max', type=float,
                        default=ny_default,
                        help='Domain top y-coordinate')
    parser.add_argument('--sigma', type=float,
                        default=1,
                        help='Normal distribution width')
    parser.add_argument('--mu_x', type=float,
                        default=nx_default/2,
                        help='Normal distribution x-mean')
    parser.add_argument('--mu_y', type=float,
                        default=ny_default/2,
                        help='Normal distribution y-mean')
    parser.add_argument('--n_bin', type=int,
                        default=10,
                        help='Number of bins')
    parser.add_argument('--min', type=int,
                        default=0,
                        help='Bin minimum')
    parser.add_argument('--max', type=int,
                        default=0.001,
                        help='Bin maximum')
    args = parser.parse_args()

    x_coords = np.linspace(args.x_min, args.x_max, args.nx + 1)
    y_coords = np.linspace(args.y_min, args.y_max, args.ny + 1)
    x_mesh, y_mesh = np.meshgrid(x_coords, y_coords)

    r2 = (x_mesh - args.mu_x) * (x_mesh - args.mu_x) \
        + (y_mesh - args.mu_y) * (y_mesh - args.mu_y)
    sigma2 = args.sigma * args.sigma
    values = np.exp(- r2 / (2 * sigma2)) / math.sqrt(2 * math.pi * sigma2)

    print(values)

    pdf = np.histogram(values, bins=args.n_bin, range=(args.min, args.max))
    print(pdf)




# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Functions to project data onto CCEW EOFs and compute activity indices.
Created by: Maria Gehne
maria.gehne@noaa.gov
2020

List of functions:

waveact:

rem_seas_cyc:

rem_seas_mean:

waveproj:

waveskill:

get_timestr:

plot_activity:

plot_skill:

plot_activity_mpl:

plot_skill_mpl:

eof_comp:

"""

import numpy as np
import xarray as xr
import plotly.graph_objects as go
from kaleido.scopes.plotly import PlotlyScope
from netCDF4 import num2date
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

colors = ['tab:blue', 'tab:orange', 'k', 'tab:green', 'tab:red', 'tab:brown']
linestyles = ('-', ':', '--')

class opt_eof():
    def __init__(self, cov):
        self.covariance = cov

def waveact(data: object, wave: str, eofpath: str, spd: int, res: str, nlat: int, tres: str,opt=False):
    """
    Main script to compute the wave activity index.
    :param data: DataArray containing the raw data
    :param wave: string describing the wave name
    :param eofpath: file path to EOFs
    :param spd: number of obs per day
    :param res: resolution of the data ('1p0')
    :param nlat: number of latitudes in the data (180 or 181)
    :param tres: temporal resolution of the eofs ('annual' or 'monthly')
    :param opt: optional parameter, currently not used
    :return: DataArray containing the time series of wave activity
    """
    # read EOFs from file
    if (wave == 'Kelvin' or wave == 'kelvin'):
        eofname = 'EOF_1-4_130-270E_-15S-15N_persiann_cdr_'+res+'_nlat'+str(nlat)+'_fillmiss8314_1983-2016_Kelvinband_'
        eofnum = np.arange(4) + 1
    elif (wave == 'ER' or wave == 'er'):
        eofname = 'EOF_1-4_60-220E_-21S-21N_persiann_cdr_'+res+'_nlat'+str(nlat)+'_fillmiss8314_1983-2016_ERband_'
        eofnum = np.arange(4) + 1
    elif (wave == 'MJO' or wave == 'mjo'):
        eofname = 'EOF_1-4_30-240E_-21S-21N_persiann_cdr_'+res+'_nlat'+str(nlat)+'_fillmiss8314_1983-2016_MJOband_'
        eofnum = np.arange(4) + 1
    elif (wave == 'MRG' or wave == 'mrg'):
        eofname = 'EOF_1-2_150-210E_-21S-21N_precip.trmm.8x.1deg.lats'+str(nlat)+'.v7a.fillmiss.comp.1998-201806_MRGband_'
        eofnum = np.arange(2) + 1

    if tres == 'monthly':
        filename = eofpath + eofname + '01.nc'
    elif tres == 'annual':
        filename = eofpath + eofname + 'annual.nc'

    ds = xr.open_dataset(filename)
    nlat = len(ds.lat)
    nlon = len(ds.lon)
    ds.close()
    month = np.arange(12) + 1
    nmon = 12
    ntim = len(data['time'])
    annual = np.arange(1)

    if tres == 'monthly':
        eofseas = xr.DataArray(0., coords=[month, eofnum, ds.lat, ds.lon], dims=['month', 'eofnum', 'lat', 'lon'])
        for ss in month:
            monthnum = f"{ss:02d}"
            ds = xr.open_dataset(eofpath + eofname + monthnum + '.nc')
            eofseas[ss - 1, :, :, :] = ds.eof
            ds.close()
    elif tres == 'annual':
        eofseas = xr.DataArray(0., coords=[annual, eofnum, ds.lat, ds.lon], dims=['annual', 'eofnum', 'lat', 'lon'])
        ds = xr.open_dataset(eofpath + eofname + 'annual.nc')
        eofseas[0, :, :, :] = ds.eof
        ds.close()

    # remove mean annual cycle
    print("remove annual cycle")
    doy = data['time.dayofyear']
    doy_oad = doy[0::spd]
    nd = len(doy_oad) - len(np.unique(doy_oad))
    if nd > 0:
        data_anom = rem_seas_cyc(data)
    else:
        data_anom = rem_seas_mean(data)

    # compute projection
    print("project onto EOFs")
    tswave = waveproj(data_anom, eofseas)

    # compute activity
    waveact = xr.DataArray(0., coords=[data_anom.time], dims=['time'])
    waveact.values = np.sqrt(np.sum(np.square(tswave), 0))
    try:
        waveact.attrs['units'] = data.attrs['units']
    except KeyError:
        print('Data has no units attribute, cannot attach units to activity')
    waveact.attrs['name'] = wave+' activity'

    del data, data_anom

    return waveact


def rem_seas_cyc(data: object, opt: object = False) -> object:
    """
    Read in a xarray data array with time coordinate containing daily data. Compute anomalies from daily climatology.
    :type data: xr.DataArray
    :param data: xarray
    :param opt: optional parameter, not currently used
    :return: xr.DataArray containing anomalies from daily climatology
    """
    da = xr.DataArray(np.arange(len(data['time'])), coords=[data['time']], dims=['time'])
    month_day_str = xr.DataArray(da.indexes['time'].strftime('%m-%d'), coords=da.coords, name='month_day_str')
    time = data['time']

    data = data.rename({'time': 'month_day_str'})
    month_day_str = month_day_str.rename({'time': 'month_day_str'})
    data = data.assign_coords(month_day_str=month_day_str)
    clim = data.groupby('month_day_str').mean('month_day_str')

    data_anom = data.groupby('month_day_str') - clim
    data_anom = data_anom.rename({'month_day_str': 'time'})
    data_anom = data_anom.assign_coords(time=time)

    return data_anom

def rem_seas_mean(data: object, opt: object = False) -> object:
    """
    Read in a xarray data array with time coordinate containing daily data. Compute anomalies from time mean.
    :type data: xr.DataArray
    :param data: xarray
    :param opt: optional parameter, not currently used
    :return: xr.DataArray containing anomalies from daily climatology
    """
    clim = data.mean('time')
    data_anom = data - clim

    return data_anom


def waveproj(data_anom: object, eofseas: object):
    """
    Compute the projection onto the CCEW EOFS
    :param data_anom: anomalies of precipitation
    :param eofseas: xarray dataarray containing the monthly EOF patterns
    :return: wave time series projected onto each EOF
    """
    data_anom = data_anom.sel(lat=slice(eofseas.lat.min(), eofseas.lat.max()),
                              lon=slice(eofseas.lon.min(), eofseas.lon.max()))
    mm = data_anom.time.dt.month
    ntim = len(data_anom.time)
    neof = len(eofseas.eofnum)
    proj_wave: object = xr.DataArray(0., coords=[eofseas.eofnum, data_anom.time], dims=['eofnum', 'time'])
    tswave: object = xr.DataArray(0., coords=[data_anom.time], dims=['time'])

    if eofseas.shape[0] == 1:
        eof = eofseas[0, :, :, :]
        for tt in range(ntim):
            for ee in range(neof):
                proj_wave[ee, tt] = eof[ee, :, :] @ data_anom[tt, :, :]
    else:
        for tt in range(ntim):
            eof = eofseas[mm[tt] - 1, :, :, :]
            for ee in range(neof):
                proj_wave[ee, tt] = eof[ee, :, :] @ data_anom[tt, :, :]

    return proj_wave


def wave_skill(act):
    """

    :param act:
    :type act:
    :return:
    :rtype:
    """
    nfchr, nlines, ntim = act.shape
    skill = act[:, 1::, 0].squeeze()

    for ff in np.arange(nfchr):
        for ll in np.arange(nlines-1):
            tmpskill = np.corrcoef(act[ff, ll+1, :], act[ff, 0, :])
            skill[ff, ll] = tmpskill[0, 1]

    return skill

def get_timestr(time):
    """
    Generate time string for y-axis labels.
    :param time: time coordinate
    :type time: datetime object
    :return: timestr
    :rtype: str
    """
    ts = (time - np.datetime64('1970-01-01T00:00:00Z')) / np.timedelta64(1, 'h')
    date = num2date(ts, 'hours since 1970-01-01T00:00:00Z')
    timestr = [i.strftime("%Y-%m-%d %H:%M") for i in date]

    return timestr


def plot_activity(act, wavename, labels, plotpath, fchr=[]):
    """
    Plot pattern correlation curves as a function of lead time.
    :param act:
    :type act:
    :param labels:
    :type labels:
    :return:
    :rtype:
    """

    plttype = "png"
    plotname = plotpath + wavename + "Activity." + plttype
    if fchr:
        plotname = plotpath + wavename + "Activity_f" + f"{fchr:03d}" + "." + plttype

    nlines, ntim = act.shape

    timestr = get_timestr(act['time'])

    colors = ['black', 'dodgerblue', 'orange', 'seagreen', 'firebrick']

    scope = PlotlyScope()
    fig = go.Figure()
    for ll in np.arange(nlines):
        fig.add_trace(go.Scatter(x=timestr, y=act[ll, :].values,
                                mode='lines',
                                name=labels[ll],
                                line=dict(color=colors[ll], width=2)))

    fig.update_layout(
        title=wavename + " FH" + f"{fchr:03d}",
        yaxis=dict(range=[0, 25]))

    #fig.update_xaxes(ticks="", tick0=0, dtick=12, title_text='date')
    fig.update_yaxes(ticks="", tick0=0, dtick=1., title_text='activity')

    with open(plotname, "wb") as f:
        f.write(scope.transform(fig, format=plttype))


def plot_skill(skill, wavename, labels, plotpath):
    """
    Plot pattern correlation curves as a function of lead time.
    :param skill:
    :type skill:
    :param labels:
    :type labels:
    :return:
    :rtype:
    """

    plttype = "png"
    plotname = plotpath + wavename + "Skill." + plttype

    nfchr, nlines = skill.shape

    colors = ['dodgerblue', 'orange', 'seagreen', 'firebrick']

    scope = PlotlyScope()
    fig = go.Figure()
    for ll in np.arange(nlines):
        fig.add_trace(go.Scatter(x=skill['fchrs'], y=skill[:, ll],
                                mode='lines',
                                name=labels[ll],
                                line=dict(color=colors[ll], width=2)))

    fig.update_layout(
        title=wavename + " skill",
        yaxis=dict(range=[0, 1]))

    fig.update_xaxes(ticks="", tick0=0, dtick=24, title_text='lead time (h)')
    fig.update_yaxes(ticks="", tick0=0, dtick=0.1, title_text='skill correlation')

    with open(plotname, "wb") as f:
        f.write(scope.transform(fig, format=plttype))


def plot_activity_mpl(act, wavename, labels, plotpath, fchr=[]):
    """
    Plot pattern correlation curves as a function of lead time.
    :param act:
    :type act:
    :param labels:
    :type labels:
    :return:
    :rtype:
    """

    plttype = "png"
    plotname = plotpath + wavename + "Activity." + plttype
    if fchr:
        plotname = plotpath + wavename + "Activity_f" + f"{fchr:03d}" + "." + plttype

    nlines, ntim = act.shape

    timestr = get_timestr(act['time'])

    fig = plt.figure(figsize=(9, 3))
    ax = fig.add_subplot(111)

    for ll in np.arange(nlines):
        ax.plot(timestr, act[ll, :].values, color=colors[ll], linestyle=linestyles[ll])
    ax.set_xlabel('time')
    ax.set_xticks(timestr[::20])
    #ax.set_xlim(0, 720)
    ax.set_ylim(0, 25)
    ax.set_ylabel('activity')
    ax.set_title(wavename + " FH" + f"{fchr:03d}")
    ax.grid()

    legendlines = []
    for ll in np.arange(nlines):
        legendlines.append( Line2D([0], [0], color=colors[ll], linestyle=linestyles[ll], lw=1) )
    ax.legend(legendlines, labels, loc='upper right', prop={'size': 8})

    plt.savefig(plotname)
    plt.close(fig)

def plot_skill_mpl(skill, wavename, labels, plotpath):
    """
    Plot pattern correlation curves as a function of lead time.
    :param skill:
    :type skill:
    :param labels:
    :type labels:
    :return:
    :rtype:
    """

    plttype = "png"
    plotname = plotpath + wavename + "Skill." + plttype

    nfchr, nlines = skill.shape

    fig = plt.figure(figsize=(9, 3))
    ax = fig.add_subplot(111)

    for ll in np.arange(nlines):
        ax.plot(skill['fchrs'], skill[:,ll], color=colors[ll], linestyle='-', linewidth=2.0)
    ax.set_xlabel('lead time (h)')
    ax.set_xticks(skill['fchrs'][::2])
    ax.set_xlim(0, 720)
    ax.set_ylim(0, 1)
    ax.set_ylabel('skill (correlation)')
    ax.set_title(wavename + " skill")
    ax.grid()

    legendlines = [Line2D([0], [0], color=colors[0], linestyle='-', lw=2),
                   Line2D([0], [0], color=colors[1], linestyle='-', lw=2)]
    ax.legend(legendlines, labels, loc='center right', prop={'size': 12})

    plt.savefig(plotname)
    plt.close(fig)


def eof_comp(data, neof, opt, dim=0):
    """
    Compute the first neof Empirical Orthogonal functions treating dim as the time dimension and
    :param data: Input data.
    :param neof: Number of EOFs to return.
    :param opt: Class parameter to change the default behavior. opt = opt_eof(True) means use the covariance matrix.
    :param dim: which dimension contains the number of observations (i.e. time dimension)
    :return: eof time series (u), singular values (s), eof patterns (vh) for the first neof EOFs
    """
    ndims = data.ndim
    dimsizes = data.shape
    dimsizesred = np.arange(ndims-1)
    dimeof = np.arange(ndims)
    diml = 1
    dd=0
    for d in np.arange(ndims):
        if d != dim:
            diml = diml * dimsizes[d]
            dimsizesred[dd] = dimsizes[d]
            dd += 1
    dimeof[0] = neof
    dimeof[1:] = dimsizesred

    # move dimension dim to 0
    if dim != 0:
        data_reord = np.moveaxis(data, dim, 0)
    else:
        data_reord = data

    # reshape to 2d by combining all dimensions 1 and up
    a = np.reshape(data_reord.values, (dimsizes[dim], diml))

    # standardize if correlation matrix instead of covariance matrix is needed
    if opt.covariance == False:
        print('Using correlation matrix for EOF computation.')
        a = (a - np.mean(a, axis=0)) / np.std(a, axis=0)
    else:
        print('Using covariance matrix for EOF computation.')

    print('Computing singular values...')
    eofts, s, eof = np.linalg.svd(a, full_matrices=False)
    print('done!')

    var = np.square(s)
    varsum = np.sum(var)
    var_exp = var/varsum

    eof = np.reshape(eof[0:neof, :], dimeof)
    return eofts[:, 0:neof], var_exp[0:neof], eof[0:neof, :]


# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: agg_stat_eqz.py

How to use:
 - Call from other Python function
        AGG_STAT_EVENT_EQZ = AggStatEventEqz(PARAMS)
        AGG_STAT_EVENT_EQZ.calculate_values()
        where PARAMS – a dictionary with data description parameters including
        location of input and output data.
        The structure is similar to Rscript template

 - Run as a stand-alone script
        python agg_stat_eqz.py <parameters_file>
        where - <parameters_file> is YAML file with parameters
        and environment variable should be set to PYTHONPATH=<path_to_METcalcpy>

 - Run from Java
        proc = Runtime.getRuntime().exec(
                “python agg_stat.eqz.py <parameters_file>”,
                new String[]{”PYTHONPATH=<path_to_METcalcpy>”},
                new File(System.getProperty("user.home")));

"""
import argparse
import sys
import logging
import pandas as pd
import yaml

from metcalcpy import GROUP_SEPARATOR
from metcalcpy.event_equalize_against_values import event_equalize_against_values
from metcalcpy.util.utils import parse_bool


class AggStatEventEqz:
    """A class that performs event equalisation logic on input data
        with MODE and MTD attribute statistics
        EE is executed against previously calculated cases

        All parameters including data description and location is in the parameters dictionary
        Usage:
            initialise this call with the parameters dictionary and then
            calls perform_ee method
            This method will execute EE and save the result to the file
                AGG_STAT_EVENT_EQZ = AggStatEventEqz(PARAMS)
                AGG_STAT_EVENT_EQZ.calculate_values()
        """

    def __init__(self, in_params):
        self.params = in_params

        self.input_data = pd.read_csv(
            self.params['agg_stat_input'],
            header=[0],
            sep='\t'
        )

        self.column_names = self.input_data.columns.values
        self.series_data = None

    def calculate_values(self):
        """Performs event equalisation if needed and saves equalized data to the file.
        """
        is_event_equal = parse_bool(self.params['event_equal'])

        # check if EE is needed
        if not self.input_data.empty and is_event_equal:
            # read previously calculated cases
            prev_cases = pd.read_csv(
                self.params['agg_stat_input_ee'],
                header=[0],
                sep='\t'
            )

            # perform for axis 1
            output_ee_data = self.perform_ee_on_axis(prev_cases, '1')

            # perform for axis 2
            if self.params['series_val_2']:
                output_ee_data = output_ee_data.append(self.perform_ee_on_axis(prev_cases, '2'))
        else:
            output_ee_data = self.input_data
            if self.input_data.empty:
                logging.info(
                    'Event equalisation was not performed because the input data is empty.'
                )

        output_ee_data.to_csv(self.params['agg_stat_output'],
                              index=None, header=True, mode='w',
                              sep="\t", na_rep="NA")

    def perform_ee_on_axis(self, prev_cases, axis='1'):
        """Performs event equalisation against previously calculated cases for the selected axis
            Returns:
                A data frame that contains equalized records
        """
        output_ee_data = pd.DataFrame()
        for fcst_var, fcst_var_stats in self.params['fcst_var_val_' + axis].items():
            for series_var, series_var_vals in self.params['series_val_' + axis].items():

                series_var_vals_no_group = []
                for val in series_var_vals:
                    split_val = val.split(GROUP_SEPARATOR)
                    series_var_vals_no_group.extend(split_val)

                # filter input data based on fcst_var, statistic and all series variables values
                series_data_for_ee = self.input_data[
                    (self.input_data['fcst_var'] == fcst_var)
                    & (self.input_data[series_var].isin(series_var_vals_no_group))
                    ]
                # filter previous cases on the same  fcst_var,
                # statistic and all series variables values
                series_data_for_prev_cases = prev_cases[
                    (prev_cases['fcst_var'] == fcst_var)
                    & (prev_cases[series_var].isin(series_var_vals_no_group))
                    ]
                # get unique cases from filtered previous cases
                series_data_for_prev_cases_unique = series_data_for_prev_cases['equalize'].unique()

                # perform ee
                series_data_after_ee = event_equalize_against_values(
                    series_data_for_ee,
                    series_data_for_prev_cases_unique)

                # append EE data to result
                if output_ee_data.empty:
                    output_ee_data = series_data_after_ee
                else:
                    output_ee_data = output_ee_data.append(series_data_after_ee)
        return output_ee_data


if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(description='List of agg_stat_event_eqz arguments')
    PARSER.add_argument("parameters_file", help="Path to YAML parameters file",
                        type=argparse.FileType('r'),
                        default=sys.stdin)
    ARGS = PARSER.parse_args()
    PARAMS = yaml.load(ARGS.parameters_file, Loader=yaml.FullLoader)

    AGG_STAT_EVENT_EQZ = AggStatEventEqz(PARAMS)
    AGG_STAT_EVENT_EQZ.calculate_values()



"""Tests the operation of calc_difficulty_index.py"""

import numpy as np
import pytest
from metcalcpy.calc_difficulty_index import forecast_difficulty

__author__ = "Lindsay Blank (NCAR)"

def test_forecast_difficulty():
    """
    Test that the output of forecast_difficulty function is correct.
    
    Returns
    -------
    None.
    """
    
    #Settings
    EPS = np.finfo(np.float32).eps
    nlon = 18
    nlat = 9
    nmembers = 10
    np.random.seed(12345)
    
    # Wave heights and wind speeds generally follow the Rayleigh distribution,
    # which is a Weibull distribution with a scale factor of 2,
    # and any shape parameter.
    # If the shape parameter is 1, then it is the same as a chi-square
    # distribution with 2 dofs.
    # Expected value E[x] = sigma * sqrt(pi/2)
    # mean_windspeed = 6.64  # meters/second
    xunits = 'feet'
    mean_height = 11.0 #mean wave height in feet
    mode_height = np.sqrt(2.0 / np.pi) * mean_height

    fieldijn = np.random.rayleigh(scale=mode_height,
            size=(nlat, nlon, nmembers))
    muij = np.mean(fieldijn, axis=-1)
    pertijn = fieldijn - np.dstack([muij] * nmembers)
    sigmaij = np.sqrt(np.mean(pertijn * pertijn, axis=-1))

    threshold = 9.0
    regularize = 0.01
    smax = 9.0
    sigma_max = smax + np.zeros_like(sigmaij)
    thresh_eps = 2.0
    kwargs = {'thresh_eps': thresh_eps, 'threshold_type': 'proximity'}

    assert 0.9095608641027515 == forecast_difficulty(sigmaij, muij, threshold, fieldijn,
            Aplin=None, sigma_over_mu_ref=EPS)[0][0]
    assert 0.8191620255148825 == forecast_difficulty(sigmaij, muij, threshold, fieldijn,
            Aplin=None, sigma_over_mu_ref=EPS)[8][17]
    assert 1.227707670365556 == forecast_difficulty(sigmaij, muij, threshold, fieldijn,
            Aplin=None, sigma_over_mu_ref=EPS)[4][9]


if __name__ == "__main__":
    test_forecast_difficulty()




# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ** University of Illinois, Urbana-Champaign
 # ============================*
 
 
 
import os
import netCDF4
import numpy as np
import datetime
from scipy import stats


def parse_steps():

    steps_param_fcst = os.environ.get('FCST_STEPS','')
    steps_list_fcst = steps_param_fcst.split("+")

    steps_param_obs = os.environ.get('OBS_STEPS','')
    steps_list_obs = steps_param_obs.split("+")

    return steps_list_fcst, steps_list_obs


def write_mpr_file(data_obs,data_fcst,lats_in,lons_in,time_obs,time_fcst,mname,desc,fvar,funit,flev,ovar,ounit,olev,maskname,obslev,outfile):

    dlength = len(lons_in)
    bdims = data_obs.shape

    index_num = np.arange(0,dlength,1)+1

    # Get the length of the model, FCST_VAR, FCST_LEV, OBS_VAR, OBS_LEV, VX_MASK
    mname_len = str(max([5,len(mname)])+3)
    desc_len = str(max([4,len(mname)])+1)
    mask_len = str(max([7,len(maskname)])+3)
    fvar_len = str(max([8,len(fvar)])+3)
    funit_len = str(max([8,len(funit)])+3)
    flev_len = str(max([8,len(flev)])+3)
    ovar_len = str(max([7,len(ovar)])+3)
    ounit_len = str(max([8,len(ounit)])+3)
    olev_len = str(max([7,len(olev)])+3)

    format_string = '%-7s %-'+mname_len+'s %-'+desc_len+'s %-12s %-18s %-18s %-12s %-17s %-17s %-'+fvar_len+'s ' \
        '%-'+funit_len+'s %-'+flev_len+'s %-'+ovar_len+'s %-'+ounit_len+'s %-'+olev_len+'s %-10s %-'+mask_len+'s ' \
        '%-13s %-13s %-13s %-13s %-13s %-13s %-9s\n'
    format_string2 = '%-7s %-'+mname_len+'s %-'+desc_len+'s %-12s %-18s %-18s %-12s %-17s %-17s %-'+fvar_len+'s ' \
        '%-'+funit_len+'s %-'+flev_len+'s %-'+ovar_len+'s %-'+ounit_len+'s %-'+olev_len+'s %-10s %-'+mask_len+'s ' \
        '%-13s %-13s %-13s %-13s %-13s %-13s %-9s %-10s %-10s %-10s %-12.4f %-12.4f %-10s %-10s %-12.4f %-12.4f ' \
        '%-10s %-10s %-10s %-10s\n'

     # Write the file
    for y in range(bdims[0]):
        for dd in range(bdims[1]):
            if time_fcst['valid'][y][dd]:
                ft_stamp = time_fcst['lead'][y][dd]+'L_'+time_fcst['valid'][y][dd][0:8]+'_' \
                    +time_fcst['valid'][y][dd][9:15]+'V'
                mpr_outfile_name = outfile+'_'+ft_stamp+'.stat'
                with open(mpr_outfile_name, 'w') as mf:
                    mf.write(format_string % ('VERSION', 'MODEL', 'DESC', 'FCST_LEAD', 'FCST_VALID_BEG', 'FCST_VALID_END',
                        'OBS_LEAD', 'OBS_VALID_BEG', 'OBS_VALID_END', 'FCST_VAR', 'FCST_UNITS', 'FCST_LEV', 'OBS_VAR', 
                        'OBS_UNITS', 'OBS_LEV', 'OBTYPE', 'VX_MASK', 'INTERP_MTHD', 'INTERP_PNTS', 'FCST_THRESH', 
                        'OBS_THRESH', 'COV_THRESH', 'ALPHA', 'LINE_TYPE'))
                    for dpt in range(dlength):
                        mf.write(format_string2 % ('V9.1',mname,desc,time_fcst['lead'][y][dd],time_fcst['valid'][y][dd],
                            time_fcst['valid'][y][dd],time_obs['lead'][y][dd],time_obs['valid'][y][dd],
                            time_obs['valid'][y][dd],fvar,funit,flev,ovar,ounit,olev,'ADPUPA',maskname,
                            'NEAREST','1','NA','NA','NA','NA','MPR',str(dlength),str(index_num[dpt]),'NA',
                            lats_in[dpt],lons_in[dpt],obslev,'NA',data_fcst[y,dd,dpt],data_obs[y,dd,dpt],'NA','NA',
                            'NA','NA'))


def read_nc_met(infiles,invar,nseasons,dseasons):

    print("Reading in Data")

    # Check to make sure that everything is not set to missing:
    if all('missing' == fn for fn in infiles):
        raise Exception('No input files found as given, check paths to input files')

    #Find the first non empty file name so I can get the variable sizes
    locin = next(sub for sub in infiles if sub != 'missing')
    indata = netCDF4.Dataset(locin)
    lats = indata.variables['lat'][:]
    lons = indata.variables['lon'][:]
    invar_arr = indata.variables[invar][:]
    indata.close()

    var_3d = np.empty([len(infiles),len(invar_arr[:,0]),len(invar_arr[0,:])])
    init_list = []
    valid_list = []
    lead_list = []

    for i in range(0,len(infiles)):

        #Read in the data
        if (infiles[i] != 'missing'):
            indata = netCDF4.Dataset(infiles[i])
            new_invar = indata.variables[invar][:]

            init_time_str = indata.variables[invar].getncattr('init_time')
            valid_time_str = indata.variables[invar].getncattr('valid_time')
            lead_dt = datetime.datetime.strptime(valid_time_str,'%Y%m%d_%H%M%S') - datetime.datetime.strptime(init_time_str,'%Y%m%d_%H%M%S')
            leadmin,leadsec = divmod(lead_dt.total_seconds(), 60)
            leadhr,leadmin = divmod(leadmin,60)
            lead_str = str(int(leadhr)).zfill(2)+str(int(leadmin)).zfill(2)+str(int(leadsec)).zfill(2)
            indata.close()
        else:
            new_invar = np.empty((1,len(var_3d[0,:,0]),len(var_3d[0,0,:])),dtype=np.float)
            init_time_str = ''
            valid_time_str = ''
            lead_str = ''
            new_invar[:] = np.nan
        init_list.append(init_time_str)
        valid_list.append(valid_time_str)
        lead_list.append(lead_str)
        var_3d[i,:,:] = new_invar

    var_4d = np.reshape(var_3d,[nseasons,dseasons,len(var_3d[0,:,0]),len(var_3d[0,0,:])])

    # Reshape time arrays and store them in a dictionary
    init_list_2d = np.reshape(init_list,[nseasons,dseasons])
    valid_list_2d = np.reshape(valid_list,[nseasons,dseasons])
    lead_list_2d = np.reshape(lead_list,[nseasons,dseasons])
    time_dict = {'init':init_list_2d,'valid':valid_list_2d,'lead':lead_list_2d}

    return var_4d,lats,lons,time_dict


def reorder_fcst_regimes(kmeans_fcst,perc_fcst,wrc_fcst,wrnum_fcst,fcst_order):

    # Reorder the data so that the weather regime patterns match between
    # the forecast and observations, is needed
    kmeans_fcst_new = np.empty(kmeans_fcst.shape,dtype=float)
    perc_fcst_new = np.empty(perc_fcst.shape,dtype=float)
    wrc_fcst_new = np.empty(wrc_fcst.shape,dtype=float)
    for wrn in np.arange(wrnum_fcst):
        perc_fcst_new[wrn] = perc_fcst[fcst_order[wrn]-1]
        kmeans_fcst_new[wrn,:,:] = kmeans_fcst[fcst_order[wrn]-1,:,:]
        wrc_cur = np.where(wrc_fcst == fcst_order[wrn])
        wrc_fcst_new[wrc_cur] = wrn + 1

    return kmeans_fcst_new,perc_fcst_new,wrc_fcst_new


def reorder_fcst_regimes_correlate(kmeans_obs,kmeans_fcst,perc_fcst,wrc_fcst,wrnum_fcst):

    # Correlate between the forecast and obs weather regimes to find the max correlation
    # Use the max correlation in reordering the forecast to match the observations
    matching_order = np.empty(wrnum_fcst,dtype=int)
    for wrr in np.arange(wrnum_fcst):
        corr_arr = np.empty(wrnum_fcst,dtype=float)
        for owr in np.arange(wrnum_fcst):
            curcorr = stats.pearsonr(kmeans_obs[wrr,:,:].flatten(),kmeans_fcst[owr,:,:].flatten())
            corr_arr[owr] = curcorr[0]
        matching_order[wrr] = corr_arr.argmax()+1

    # Call the reorder_fcst_regimes to reorder
    matching_order = list(matching_order)
    kmeans_fcst_new,perc_fcst_new,wrc_fcst_new = reorder_fcst_regimes(kmeans_fcst,perc_fcst,wrc_fcst,wrnum_fcst,matching_order)

    return kmeans_fcst_new,perc_fcst_new,wrc_fcst_new 



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: sal1l2_statistics.py
"""
import warnings
import numpy as np
from metcalcpy.util.utils import round_half_up, sum_column_data_by_name, PRECISION, get_total_values

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'


def calculate_anom_corr(input_data, columns_names, aggregation=False):
    """Performs calculation of ANOM_CORR - The Anomaly Correlation
     including normal confidence limits

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ANOM_CORR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        ffbar = sum_column_data_by_name(input_data, columns_names, 'ffabar') / total
        fbar = sum_column_data_by_name(input_data, columns_names, 'fabar') / total
        oobar = sum_column_data_by_name(input_data, columns_names, 'ooabar') / total
        obar = sum_column_data_by_name(input_data, columns_names, 'oabar') / total
        fobar = sum_column_data_by_name(input_data, columns_names, 'foabar') / total
        v = (total ** 2 * ffbar - total ** 2 * fbar ** 2) \
            * (total ** 2 * oobar - total ** 2 * obar ** 2)
        if v <= 0:
            return None
        anom_corr = (total ** 2 * fobar - total ** 2 * fbar * obar) / np.sqrt(v)
        if anom_corr > 1:
            anom_corr = None
        else:
            anom_corr = round_half_up(anom_corr, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        anom_corr = None
    warnings.filterwarnings('ignore')
    return anom_corr


def calculate_anom_corr_raw(input_data, columns_names, aggregation=False):
    """Performs calculation of ANOM_CORR_RAW - The Uncentered Anomaly Correlation
     including normal confidence limits

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated ANOM_CORR_RAW as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        ffbar = sum_column_data_by_name(input_data, columns_names, 'ffabar') / total
        oobar = sum_column_data_by_name(input_data, columns_names, 'ooabar') / total
        fobar = sum_column_data_by_name(input_data, columns_names, 'foabar') / total
        v = ffbar * oobar
        if v < 0:
            return None
        anom_corr_raw = fobar / np.sqrt(v)
        if anom_corr_raw > 1:
            anom_corr_raw = 1
        elif anom_corr_raw < -1:
            anom_corr_raw = -1

        anom_corr_raw = round_half_up(anom_corr_raw, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        anom_corr_raw = None
    warnings.filterwarnings('ignore')
    return anom_corr_raw


def calculate_rmsfa(input_data, columns_names, aggregation=False):
    """Performs calculation of RMSFA - Root mean squared forecast anomaly (f-c)
     including normal confidence limits

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated RMSFA as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        ffbar = sum_column_data_by_name(input_data, columns_names, 'ffabar') / total
        if ffbar is None or ffbar < 0:
            result = None
        else:
            result = np.sqrt(ffbar)
            result = round_half_up(result, PRECISION)
    except (TypeError, Warning):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_rmsoa(input_data, columns_names, aggregation=False):
    """Performs calculation of RMSOA - Root mean squared observation anomaly (o-c)
     including normal confidence limits

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated RMSOA as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        oobar = sum_column_data_by_name(input_data, columns_names, 'ooabar') / total
        if oobar is None or oobar < 0:
            result = None
        else:
            result = np.sqrt(oobar)
            result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result



"""Tests the operation of METcalcpy's utils code."""
import numpy as np
import pandas as pd
import pytest

from metcalcpy.util.utils import represents_int, is_string_integer, get_derived_curve_name, calc_derived_curve_value, \
    unique, intersection, is_derived_point, parse_bool, round_half_up, sum_column_data_by_name, \
    nrow_column_data_by_name_value, create_permutations_mv, column_data_by_name, calculate_mtd_revision_stats, \
    autocor_coef, is_string_strictly_float


@pytest.fixture
def settings():
    """Initialise values for testing.

    Returns:
        dictionary with values of different type
    """
    settings_dict = dict()
    columns = np.array(['model', 'fcst_init_beg', 'fcst_valid_beg', 'fcst_lead', 'vx_mask', 'fcst_var',
                        'stat_name', 'stat_value', 'total', 'fbar', 'obar', 'fobar', 'ffbar', 'oobar',
                        'mae'])
    settings_dict['columns'] = columns

    return settings_dict


def test_represents_int():
    assert represents_int(1)
    assert not represents_int("1")
    assert not represents_int(1.5)
    assert not represents_int(None)


def test_is_string_integer():
    assert is_string_integer("1")
    assert not is_string_integer("1.5")
    assert not is_string_integer("not_int")
    assert not is_string_integer(None)


def test_is_string_float():
    assert not is_string_strictly_float("1")
    assert is_string_strictly_float("1.5")
    assert not is_string_strictly_float("not_float")
    assert not is_string_strictly_float("not.float")
    assert not is_string_strictly_float(None)


def test_get_derived_curve_name():
    list_of_names = ['analog_e GHI FBAR', 'dicast15 GHI FBAR']
    assert 'DIFF(analog_e GHI FBAR-dicast15 GHI FBAR)' == get_derived_curve_name(list_of_names)
    list_of_names.append('RATIO')
    assert 'RATIO(analog_e GHI FBAR/dicast15 GHI FBAR)' == get_derived_curve_name(list_of_names)


def test_calc_derived_curve_value():
    val1 = np.array([674.08, 100])
    val2 = np.array([665.97, 100])
    result = calc_derived_curve_value(val1, val2, 'DIFF')
    assert np.allclose(np.array([8.11, 0.]), result)
    result = calc_derived_curve_value(val1, val2, 'RATIO')
    assert np.allclose(np.array([1.01217773, 1.]), result)
    val2[1] = 0
    result = calc_derived_curve_value(val1, val2, 'RATIO')
    assert not result


def test_unique():
    in_list = ['analog_e', 'GHI', 'FBAR', '3000', 'FBAR', '3000']
    out_list = unique(in_list)
    out_list.sort()
    assert ['3000', 'FBAR', 'GHI', 'analog_e'] == out_list
    assert not unique(None)


def test_intersection():
    l_1 = ['a', 'b', 'c', 'd']
    l_2 = ['g', 'b', 'o', 'd']
    assert ['b', 'd'] == intersection(l_1, l_2)
    assert not intersection(None, l_2)


def test_is_derived_point():
    point = ('stochmp1', '20000', 'PSTD_BRIER')
    assert not is_derived_point(point)
    point = (
        'DIFF(stochmp1 TMP_ENS_FREQ_ge283 PSTD_BRIER-stochmp2 TMP_ENS_FREQ_ge283 PSTD_BRIER)', '20000', 'PSTD_BRIER')
    assert is_derived_point(point)


def test_parse_bool():
    assert parse_bool("True")
    assert not parse_bool("False")


def test_round_half_up():
    assert 1.2 == round_half_up(1.23, 1)
    assert 1.3 == round_half_up(1.28, 1)
    assert 1.3 == round_half_up(1.25, 1)
    assert -1.2 == round_half_up(-1.25, 1)
    assert 0.12346 == round_half_up(0.1234567875, 5)
    assert 456792.12346 == round_half_up(456792.1234567875, 5)


def test_sum_column_data_by_name(settings):
    data_values = np.array([
        ['dicast15', '2019-07-03 12:00:00', '2019-07-04 09:00:00', 210000, 'SWS01', 'GHI', 'MAE', 0, 1, 1073.4, 1085.7,
         1165390.38, 1152187.56, 1178744.49, 12.3],
        ['dicast15', '2019-07-03 12:00:00', '2019-07-05 13:15:00', 491500, 'SWS01', 'GHI', 'MAE', 0, 1, 518.43, 501.36,
         259920.0648, 268769.6649, 251361.8496, 17.07]
    ])
    column_name = 'fobar'
    assert 1425310.4448 == sum_column_data_by_name(data_values, settings['columns'], column_name)

    column_name = 'not_in_array'
    assert not sum_column_data_by_name(data_values, settings['columns'], column_name)

    data_values = np.append(data_values,
                            [['dicast15', '2019-07-03 12:00:00', '2019-07-04 09:00:00', 210000, 'SWS01', 'GHI', 'MAE',
                              0, 1, 1073.4, 1085.7, None, 1152187.56, 1178744.49, 12.3]], axis=0)
    column_name = 'fobar'
    assert not sum_column_data_by_name(data_values, settings['columns'], column_name, rm_none=False)
    assert 1425310.4448 == sum_column_data_by_name(data_values, settings['columns'], column_name, rm_none=True)


def test_column_data_by_name(settings):
    data_values = np.array([
        ['dicast15', '2019-07-03 12:00:00', '2019-07-04 09:00:00', 210000, 'SWS01', 'GHI', 'MAE', 0, 1, 1073.4, 1085.7,
         1165390.38, 1152187.56, 1178744.49, 12.3],
        ['dicast15', '2019-07-03 12:00:00', '2019-07-05 13:15:00', 491500, 'SWS01', 'GHI', 'MAE', 0, 1, 518.43, 501.36,
         259920.0648, 268769.6649, 251361.8496, 17.07]
    ])
    column_name = 'fobar'
    result = column_data_by_name(data_values, settings['columns'], column_name)
    assert len(result) == 2
    assert float(data_values[0, 11]) == result[0]
    assert float(data_values[1, 11]) == result[1]

    column_name = 'not_in_array'
    assert not column_data_by_name(data_values, settings['columns'], column_name)

    data_values = np.append(data_values,
                            [['dicast15', '2019-07-03 12:00:00', '2019-07-04 09:00:00', 210000, 'SWS01', 'GHI', 'MAE',
                              0, 1, 1073.4, 1085.7, None, 1152187.56, 1178744.49, 12.3]], axis=0)
    column_name = 'fobar'
    result = column_data_by_name(data_values, settings['columns'], column_name)
    assert len(result) == 3
    assert np.isnan(result[2])

    result = column_data_by_name(data_values, settings['columns'], column_name, rm_none=True)
    assert len(result) == 2


def test_nrow_column_data_by_name_value(settings):
    data_values = np.array([
        ['dicast15', '2019-07-03 12:00:00', '2019-07-04 09:00:00', 210000, 'SWS01', 'GHI', 'MAE', 0, 1, 1073.4, 1085.7,
         1165390.38, 1152187.56, 1178744.49, 12.3],
        ['dicast15', '2019-07-03 12:00:00', '2019-07-05 13:15:00', 491500, 'SWS01', 'GHI', 'MAE', 0, 1, 518.43, 501.36,
         259920.0648, 268769.6649, 251361.8496, 17.07],
        ['dicast15', '2019-07-03 12:00:00', '2019-07-05 13:15:00', 491500, 'SWS01', 'GHI', 'MAE', 0, 1, 518.43, 501.36,
         259920.0648, 268769.6649, 251361.8496, 17.07],
        ['dicast15', '2019-07-03 12:00:00', '2019-07-05 13:15:00', 491500, 'SWS01', 'GHI', 'MAE', 1, 1, 518.43, 501.36,
         259920.0648, 268769.6649, 251361.8496, 17.07]
    ])
    filters = {'fcst_lead': 491500, 'stat_value': 0}
    assert 2 == nrow_column_data_by_name_value(data_values, settings['columns'], filters)


def test_create_permutations_mv_dict():
    fields_values = {'vx_mask': ['FULL'], 'model': ['HREF', 'HREFV3'], 'fcst_var': ['HGT'],
                     'stat_name': ['ECNT_RMSE', 'ECNT_SPREAD']}
    expected_result = [['FULL', 'HREF', 'HGT', 'ECNT_RMSE'], ['FULL', 'HREFV3', 'HGT', 'ECNT_RMSE'],
                       ['FULL', 'HREF', 'HGT', 'ECNT_SPREAD'], ['FULL', 'HREFV3', 'HGT', 'ECNT_SPREAD']]
    result = create_permutations_mv(fields_values, 0)
    assert expected_result == result


def test_create_permutations_mv_dict_empty():
    fields_values = {'vx_mask': ['FULL'], 'model': ['HREF', 'HREFV3'], 'fcst_var': [],
                     'stat_name': ['ECNT_RMSE', 'ECNT_SPREAD']}
    expected_result = [['FULL', 'HREF', 'ECNT_RMSE'], ['FULL', 'HREFV3', 'ECNT_RMSE'],
                       ['FULL', 'HREF', 'ECNT_SPREAD'], ['FULL', 'HREFV3', 'ECNT_SPREAD']]
    result = create_permutations_mv(fields_values, 0)
    assert expected_result == result


def test_create_permutations_mv_list():
    fields_values = [['FULL'], ['HREF', 'HREFV3'], ['HGT'], ['ECNT_RMSE', 'ECNT_SPREAD']]
    expected_result = [['FULL', 'HREF', 'HGT', 'ECNT_RMSE'], ['FULL', 'HREFV3', 'HGT', 'ECNT_RMSE'],
                       ['FULL', 'HREF', 'HGT', 'ECNT_SPREAD'], ['FULL', 'HREFV3', 'HGT', 'ECNT_SPREAD']]
    result = create_permutations_mv(fields_values, 0)
    assert expected_result == result


def test_calculate_mtd_revision_stats():
    df = pd.read_csv('data/mtd_revision.data', index_col=0)
    stats = calculate_mtd_revision_stats(df)
    assert stats.get("ww_run") == 0
    assert stats.get("auto_cor_p") == 0
    assert stats.get("auto_cor_r") == 0.01
    stats = calculate_mtd_revision_stats(df, 1)
    assert stats.get("auto_cor_r") == 0.15

def test_autocor_coef():
    z = [28, 28, 26, 19, 16, 24, 26, 24, 24, 29, 29, 27, 31, 26, 38, 23, 13, 14, 28, 19, 19, 17, 22, 2, 4, 5, 7, 8, 14, 14, 23]
    result = autocor_coef(z)
    assert round_half_up(result, 7) == 0.6635721

if __name__ == "__main__":
    test_represents_int()
    test_is_string_integer()
    test_is_string_float()
    test_get_derived_curve_name()
    test_calc_derived_curve_value()
    test_unique()
    test_intersection()
    test_is_derived_point()
    test_parse_bool()
    test_round_half_up()
    test_sum_column_data_by_name()
    test_calculate_mtd_revision_stats()
    test_autocor_coef()



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: ssvar_statistics.py
"""
import warnings
import numpy as np

from metcalcpy.util.sal1l2_statistics import calculate_anom_corr
from metcalcpy.util.sl1l2_statistics import calculate_fbar, calculate_fstdev, \
    calculate_obar, calculate_ostdev, \
    calculate_pr_corr, calculate_me, calculate_estdev, calculate_mse, \
    calculate_bcmse, calculate_bcrmse, calculate_rmse, \
    calculate_me2, calculate_msess
from metcalcpy.util.utils import round_half_up, sum_column_data_by_name, PRECISION, get_total_values

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'


def calculate_ssvar_fbar(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_FBAR - Average forecast value

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_FBAR as float
            or None if some of the data values are missing or invalid
    """
    return calculate_fbar(input_data, columns_names, aggregation)


def calculate_ssvar_fstdev(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_FSTDEV - Standard deviation of the error

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_FSTDEV as float
            or None if some of the data values are missing or invalid
    """
    return calculate_fstdev(input_data, columns_names, aggregation)


def calculate_ssvar_obar(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_OBAR - Average observed value

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_OBAR as float
            or None if some of the data values are missing or invalid
    """
    return calculate_obar(input_data, columns_names, aggregation)


def calculate_ssvar_ostdev(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_OSTDEV - Standard deviation of the error

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_OSTDEV as float
            or None if some of the data values are missing or invalid
    """
    return calculate_ostdev(input_data, columns_names, aggregation)


def calculate_ssvar_pr_corr(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_PR_CORR - Pearson correlation coefficient

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_PR_CORR as float
            or None if some of the data values are missing or invalid
    """
    return calculate_pr_corr(input_data, columns_names, aggregation)


def calculate_ssvar_me(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_ME - Mean error

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_ME as float
            or None if some of the data values are missing or invalid
    """
    return calculate_me(input_data, columns_names, aggregation)


def calculate_ssvar_estdev(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_ESTDEV - Standard deviation of the error

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_ESTDEV as float
            or None if some of the data values are missing or invalid
    """
    return calculate_estdev(input_data, columns_names, aggregation)


def calculate_ssvar_mse(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_MSE - Mean squared error

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated SSVAR_MSE as float
            or None if some of the data values are missing or invalid
    """
    return calculate_mse(input_data, columns_names, aggregation)


def calculate_ssvar_bcmse(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_BCMSE - Bias corrected root mean squared error

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_BCMSE as float
            or None if some of the data values are missing or invalid
    """
    return calculate_bcmse(input_data, columns_names, aggregation)


def calculate_ssvar_bcrmse(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_BCRMSE - Bias corrected root mean squared error

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_BCRMSE as float
            or None if some of the data values are missing or invalid
    """
    return calculate_bcrmse(input_data, columns_names, aggregation)


def calculate_ssvar_rmse(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_RMSE - Root mean squared error

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_RMSE as float
            or None if some of the data values are missing or invalid
    """
    return calculate_rmse(input_data, columns_names, aggregation)


def calculate_ssvar_anom_corr(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_ANOM_CORR -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_ANOM_CORR as float
            or None if some of the data values are missing or invalid
    """
    # change the names to comply with sal1l2 names
    sal1l2_columns_names = np.copy(columns_names)
    sal1l2_columns_names[sal1l2_columns_names  == 'ffbar'] = 'ffabar'
    sal1l2_columns_names[sal1l2_columns_names  == 'fbar'] = 'fabar'
    sal1l2_columns_names[sal1l2_columns_names  == 'oobar'] = 'ooabar'
    sal1l2_columns_names[sal1l2_columns_names  == 'obar'] = 'oabar'
    sal1l2_columns_names[sal1l2_columns_names  == 'fobar'] = 'foabar'
    return calculate_anom_corr(input_data, sal1l2_columns_names, aggregation)


def calculate_ssvar_me2(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_ME2 -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_ME2 as float
            or None if some of the data values are missing or invalid
    """
    return calculate_me2(input_data, columns_names, aggregation)


def calculate_ssvar_msess(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_MSESS -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_MSESS as float
            or None if some of the data values are missing or invalid
    """

    return calculate_msess(input_data, columns_names, aggregation)


def calculate_ssvar_spread(input_data, columns_names, aggregation=False):
    """Performs calculation of SSVAR_SPREAD -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated SSVAR_SPREAD as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        var_mean = sum_column_data_by_name(input_data, columns_names, 'var_mean') / total
        result = np.sqrt(var_mean)
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ssvar_total(input_data, columns_names):
    """Performs calculation of Total number of matched pairs for
        Spread/Skill Variance
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Total number of matched pairs as float
            or None if some of the data values are missing or invalid
    """
    total = sum_column_data_by_name(input_data, columns_names, 'total_orig')
    return round_half_up(total, PRECISION)



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: grad_statistics.py
"""
import warnings
from metcalcpy.util.utils import round_half_up, sum_column_data_by_name, PRECISION, get_total_values

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'


def calculate_fgbar(input_data, columns_names, aggregation=False):
    """Performs calculation of FGBAR - Mean of absolute value of forecast gradients

            Args:
                input_data: 2-dimensional numpy array with data for the calculation
                    1st dimension - the row of data frame
                    2nd dimension - the column of data frame
                columns_names: names of the columns for the 2nd dimension as Numpy array
                aggregation: if the aggregation on fields was performed

            Returns:
                calculated FGBAR as float
                or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        result = sum_column_data_by_name(input_data, columns_names, 'fgbar') / total
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ogbar(input_data, columns_names, aggregation=False):
    """Performs calculation of OGBAR - Mean of absolute value of observed gradients

            Args:
                input_data: 2-dimensional numpy array with data for the calculation
                    1st dimension - the row of data frame
                    2nd dimension - the column of data frame
                columns_names: names of the columns for the 2nd dimension as Numpy array
                aggregation: if the aggregation on fields was performed

            Returns:
                calculated OGBAR as float
                or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        result = sum_column_data_by_name(input_data, columns_names, 'ogbar') / total
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_mgbar(input_data, columns_names, aggregation=False):
    """Performs calculation of MGBAR - Mean of maximum of absolute values
        of forecast and observed gradients

            Args:
                input_data: 2-dimensional numpy array with data for the calculation
                    1st dimension - the row of data frame
                    2nd dimension - the column of data frame
                columns_names: names of the columns for the 2nd dimension as Numpy array
                aggregation: if the aggregation on fields was performed

            Returns:
                calculated MGBAR as float
                or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        result = sum_column_data_by_name(input_data, columns_names, 'mgbar') / total
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_egbar(input_data, columns_names, aggregation=False):
    """Performs calculation of EGBAR - Mean of absolute value of forecast minus observed gradients

            Args:
                input_data: 2-dimensional numpy array with data for the calculation
                    1st dimension - the row of data frame
                    2nd dimension - the column of data frame
                columns_names: names of the columns for the 2nd dimension as Numpy array
                aggregation: if the aggregation on fields was performed

            Returns:
                calculated EGBAR as float
                or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        result = sum_column_data_by_name(input_data, columns_names, 'egbar') / total
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_s1(input_data, columns_names, aggregation=False):
    """Performs calculation of S1 - S1 score

            Args:
                input_data: 2-dimensional numpy array with data for the calculation
                    1st dimension - the row of data frame
                    2nd dimension - the column of data frame
                columns_names: names of the columns for the 2nd dimension as Numpy array
                aggregation: if the aggregation on fields was performed

            Returns:
                calculated S1 as float
                or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        egbar = sum_column_data_by_name(input_data, columns_names, 'egbar') / total
        mgbar = sum_column_data_by_name(input_data, columns_names, 'mgbar') / total
        result = 100 * egbar / mgbar
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_s1_og(input_data, columns_names, aggregation=False):
    """Performs calculation of S1_OG - S1 score with respect to observed gradient

            Args:
                input_data: 2-dimensional numpy array with data for the calculation
                    1st dimension - the row of data frame
                    2nd dimension - the column of data frame
                columns_names: names of the columns for the 2nd dimension as Numpy array
                aggregation: if the aggregation on fields was performed

            Returns:
                calculated S1_OG as float
                or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        egbar = sum_column_data_by_name(input_data, columns_names, 'egbar') / total
        ogbar = sum_column_data_by_name(input_data, columns_names, 'ogbar') / total
        result = 100 * egbar / ogbar
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_fgog_ratio(input_data, columns_names, aggregation=False):
    """Performs calculation of FGOG_RATIO - Ratio of forecast and observed gradients

            Args:
                input_data: 2-dimensional numpy array with data for the calculation
                    1st dimension - the row of data frame
                    2nd dimension - the column of data frame
                columns_names: names of the columns for the 2nd dimension as Numpy array
                aggregation: if the aggregation on fields was performed

            Returns:
                calculated FGOG_RATIO as float
                or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        fgbar = sum_column_data_by_name(input_data, columns_names, 'fgbar') / total
        ogbar = sum_column_data_by_name(input_data, columns_names, 'ogbar') / total
        result = 100 * fgbar / ogbar
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_grad_total(input_data, columns_names):
    """Performs calculation of Total number of matched pairs for Gradient partial sums
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Total number of matched pairs as float
            or None if some of the data values are missing or invalid
    """
    total = sum_column_data_by_name(input_data, columns_names, 'total')
    return round_half_up(total, PRECISION)



import pytest

from metcalcpy.compare_images import CompareImages


@pytest.fixture
def settings():
    compare_diff = CompareImages('data/img_1.png', 'data/img_2.png')
    compare_same = CompareImages('data/img_1.png', 'data/img_1.png')
    settings_dict = dict()
    settings_dict['compare_diff'] = compare_diff
    settings_dict['compare_same'] = compare_same
    return settings_dict


def test_get_ssim(settings):
    assert settings['compare_diff'].get_mssim() != 1.0
    assert settings['compare_same'].get_mssim() == 1.0


def test_save_difference_image(settings):
    settings['compare_diff'].save_difference_image('data/img_diff.png')



import datetime as dt
import pytest

from metcalcpy.validate_mv_python import get_testing_period, replace_name


def test_get_testing_period():
    param = {'start_date': '2020-01-01', 'end_date': '2020-01-16'}
    (start, end) = get_testing_period(param)
    assert start == dt.datetime.strptime('2020-01-01', '%Y-%m-%d').date()
    assert end == dt.datetime.strptime('2020-01-16', '%Y-%m-%d').date()

    param = {}
    (start, end) = get_testing_period(param)
    assert start == (dt.datetime.now() - dt.timedelta(1)).date()
    assert end == dt.datetime.now().date()

    param = {'start_date': '2020-01-01'}
    (start, end) = get_testing_period(param)
    assert start == dt.datetime.strptime('2020-01-01', '%Y-%m-%d').date()
    assert end == dt.datetime.now().date()

    param = {'end_date': '2020-01-01'}
    with pytest.raises(Exception):
        get_testing_period(param)


def test_replace_name():
    assert replace_name('plot_20200115_135714.xml', 'py') == 'plot_20200115_135714_py.xml'



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: bootstrap.py
"""

import numpy as _np
from collections import Iterable
import multiprocessing as _multiprocessing
import scipy.sparse as _sparse

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'


class BootstrapResults(object):
    def __init__(self, lower_bound, value, upper_bound):
        try:
            self.lower_bound = lower_bound
            self.upper_bound = upper_bound
            self.value = value
            if self.lower_bound > self.upper_bound:
                self.lower_bound, self.upper_bound = self.upper_bound, self.lower_bound
        except TypeError:
            pass
        self.values = None
        self.distributions = None

    def __str__(self):
        return '{1}    ({0}, {2})'.format(self.lower_bound, self.value,
                                          self.upper_bound)

    def __repr__(self):
        return self.__str__()

    def _apply(self, other, func):
        return BootstrapResults(func(self.lower_bound, other),
                                func(self.value, other),
                                func(self.upper_bound, other))

    def __add__(self, other):
        return self._apply(float(other), lambda x, other: other + x)

    def __radd__(self, other):
        return self._apply(float(other), lambda x, other: other + x)

    def __sub__(self, other):
        return self._apply(float(other), lambda x, other: x - other)

    def __rsub__(self, other):
        return self._apply(float(other), lambda x, other: other - x)

    def __mul__(self, other):
        return self._apply(float(other), lambda x, other: x * other)

    def __rmul__(self, other):
        return self._apply(float(other), lambda x, other: x * other)

    def error_width(self):
        '''Returns: upper_bound - lower_bound'''
        return self.upper_bound - self.lower_bound

    def error_fraction(self):
        '''Returns the error_width / value'''
        if self.value == 0:
            return _np.inf
        else:
            return self.error_width() / self.value

    def is_significant(self):
        return _np.sign(self.upper_bound) == _np.sign(self.lower_bound)

    def get_result(self):
        '''Returns:
            -1 if statistically significantly negative
            +1 if statistically significantly positive
            0 otherwise
        '''
        return int(self.is_significant()) * _np.sign(self.value)

    def set_original_values(self, values):
        """Sets values to the original data array
            Args: values - numpy array
        """
        self.values = values

    def set_distributions(self, distributions):
        """Sets distributions y
            Args: distributions - numpy array
        """
        self.distributions = distributions


def bootstrap_and_value(values, stat_func, alpha=0.05,
                        num_iterations=1000, iteration_batch_size=None,
                        num_threads=1, ci_method='perc',
                        save_data=True, save_distributions=False, block_length: int = 1, eclv: bool = False):
    """Returns bootstrap estimate. Can do the independent and identically distributed (IID)
        or Circular Block Bootstrap (CBB) methods depending on the block_length
        Args:
            values: numpy array  of values to bootstrap
            stat_func: statistic to bootstrap. We provide several default functions:
                    * stat_functions.mean
                    * stat_functions.sum
                    * stat_functions.std
            alpha: alpha value representing the confidence interval.
                Defaults to 0.05, i.e., 95th-CI.
            num_iterations: number of bootstrap iterations to run. The higher this
                number the more sure you can be about the stability your bootstrap.
                By this - we mean the returned interval should be consistent across
                runs for the same input. This also consumes more memory and makes
                analysis slower. Defaults to 10000.
            iteration_batch_size: The bootstrap sample can generate very large
                matrices. This argument limits the memory footprint by
                batching bootstrap rounds. If unspecified the underlying code
                will produce a matrix of len(values) x num_iterations. If specified
                the code will produce sets of len(values) x iteration_batch_size
                (one at a time) until num_iterations have been simulated.
                Defaults to no batching.

            num_threads: The number of therads to use. This speeds up calculation of
                the bootstrap. Defaults to 1. If -1 is specified then
                multiprocessing.cpu_count() is used instead.
            ci_method: method for bootstrapping confidence intervals.
            save_data: Save or not the original data to the resulting object
            save_distributions: Save or not the distributions to the resulting object
            block_length: number giving the desired block lengths.
                Default (block.length = 1) is to do IID resamples.
                Should be longer than the length of dependence in the data,
                but much shorter than the size of the data. Generally, the square
                root of the sample size is a good choice
             eclv: indicates if this bootstrap estimate is for the Economic Cost Loss Relative Value or not
                Default (eclv = false)
        Returns:
            BootstrapDistributionResults representing CI, stat value and the original distribution.
    """

    values_lists = [values]
    stat_func_lists = [stat_func]

    def do_division(distr):
        return distr

    stat_val = stat_func(values)[0]
    sz = num_iterations / block_length
    distributions = _bootstrap_distribution_cbb(values_lists,
                                                stat_func_lists,
                                                num_iterations,
                                                iteration_batch_size,
                                                num_threads, block_length)

    bootstrap_dist = do_division(*distributions)
    if eclv:
        result = _get_confidence_interval_and_value_eclv(bootstrap_dist, stat_val, alpha, ci_method)
    else:
        result = _get_confidence_interval_and_value(bootstrap_dist, stat_val, alpha, ci_method)
    if save_data:
        result.set_original_values(values)
    if save_distributions:
        result.set_distributions(bootstrap_dist.flatten('F'))
    return result


def _bootstrap_distribution_cbb(values_lists, stat_func_lists,
                                num_iterations, iteration_batch_size, num_threads, block_length=1):
    '''Returns the simulated bootstrap distribution. The idea is to sample the same
        indexes in a bootstrap re-sample across all arrays passed into values_lists.

        This is especially useful when you want to co-sample records in a ratio metric.
            numerator[k].sum() / denominator[k].sum()
        and not
            numerator[ j ].sum() / denominator[k].sum()
    Args:
        values_lists: list of numpy arrays
            each represents a set of values to bootstrap. All arrays in values_lists
            must be of the same length.
        stat_func_lists: statistic to bootstrap for each element in values_lists.
        num_iterations: number of bootstrap iterations / resamples / simulations
            to perform.
        iteration_batch_size: The bootstrap sample can generate very large
            matrices. This argument limits the memory footprint by
            batching bootstrap rounds. If unspecified the underlying code
            will produce a matrix of len(values) x num_iterations. If specified
            the code will produce sets of len(values) x iteration_batch_size
            (one at a time) until num_iterations have been simulated.
            Defaults to no batching.
        num_threads: The number of therads to use. This speeds up calculation of
            the bootstrap. Defaults to 1. If -1 is specified then
            multiprocessing.cpu_count() is used instead.
        block_length: number giving the desired block lengths.
                Default (block.length = 1) is to do IID resamples.
                Should be longer than the length of dependence in the data,
                but much shorter than the size of the data.Generally, the square
                root of the sample size is a good choice
    Returns:
        The set of bootstrap resamples where each stat_function is applied on
        the bootsrapped values.
    '''

    _validate_arrays(values_lists)

    if iteration_batch_size is None:
        iteration_batch_size = num_iterations

    num_iterations = int(num_iterations)
    iteration_batch_size = int(iteration_batch_size)

    num_threads = int(num_threads)

    if num_threads == -1:
        num_threads = _multiprocessing.cpu_count()

    if num_threads <= 1:
        results = _bootstrap_sim_cbb(values_lists, stat_func_lists,
                                     num_iterations, iteration_batch_size, None, block_length)
    else:
        pool = _multiprocessing.Pool(num_threads)

        iter_per_job = _np.ceil(num_iterations * 1.0 / num_threads)

        results = []
        for seed in _np.random.randint(0, 2 ** 32 - 1, num_threads):
            r = pool.apply_async(_bootstrap_sim_cbb, (values_lists, stat_func_lists,
                                                      iter_per_job,
                                                      iteration_batch_size, seed, block_length))
            results.append(r)

        results = _np.hstack([res.get() for res in results])

        pool.close()

    return results


def bootstrap_and_value_mode(values, cases, stat_func, alpha=0.05,
                             num_iterations=1000, iteration_batch_size=None,
                             num_threads=1, ci_method='perc',
                             save_data=True, save_distributions=False, block_length=1):
    """Returns bootstrap estimate.
        Args:
            values: numpy array  of values to bootstrap
            stat_func: statistic to bootstrap. We provide several default functions:
                    * stat_functions.mean
                    * stat_functions.sum
                    * stat_functions.std
            alpha: alpha value representing the confidence interval.
                Defaults to 0.05, i.e., 95th-CI.
            num_iterations: number of bootstrap iterations to run. The higher this
                number the more sure you can be about the stability your bootstrap.
                By this - we mean the returned interval should be consistent across
                runs for the same input. This also consumes more memory and makes
                analysis slower. Defaults to 10000.
            iteration_batch_size: The bootstrap sample can generate very large
                matrices. This argument limits the memory footprint by
                batching bootstrap rounds. If unspecified the underlying code
                will produce a matrix of len(values) x num_iterations. If specified
                the code will produce sets of len(values) x iteration_batch_size
                (one at a time) until num_iterations have been simulated.
                Defaults to no batching.

            num_threads: The number of therads to use. This speeds up calculation of
                the bootstrap. Defaults to 1. If -1 is specified then
                multiprocessing.cpu_count() is used instead.
            ci_method: method for bootstrapping confidence intervals.
            save_data: Save or not the original data to the resulting object
            save_distributions: Save or not the distributions
            block_length: number giving the desired block lengths.
                Default (block.length = 1) is to do IID resamples.
                Should be longer than the length of dependence in the data,
                but much shorter than the size of the data. Generally, the square
                root of the sample size is a good choice
        Returns:
            BootstrapDistributionResults representing CI, stat value and the original distribution.
    """

    values_lists = [cases]

    stat_func_lists = [stat_func]

    def do_division(distr):
        return distr

    data_cases = _np.asarray(values['case'])
    flat_cases = cases.flatten()
    values_current = values[_np.in1d(data_cases, flat_cases)].to_numpy()
    stat_val = stat_func(values_current)[0]
    distributions = _bootstrap_distribution_cbb(values_lists,
                                                stat_func_lists,
                                                num_iterations,
                                                iteration_batch_size,
                                                num_threads, block_length)

    bootstrap_dist = do_division(*distributions)
    result = _get_confidence_interval_and_value(bootstrap_dist, stat_val, alpha, ci_method)
    if save_data:
        result.set_original_values(values)
    if save_distributions:
        result.set_distributions(bootstrap_dist.flatten('F'))
    return result


def _get_confidence_interval_and_value(bootstrap_dist, stat_val, alpha, ci_method):
    """Get the bootstrap confidence interval for a given distribution.
        Args:
            bootstrap_dist: numpy array of bootstrap results from
                bootstrap_distribution() or  bootstrap_distribution_cbb()
                or bootstrap_ab_distribution()
            stat_val: The overall statistic that this method is attempting to
                calculate error bars for.
            alpha: The alpha value for the confidence intervals.
            ci_method: if true, use the pivotal method. if false, use the
                percentile method.
    """

    # TODO Only percentile method for the confident intervals is implemented

    if stat_val is None:
        ci_method = "None"

    if ci_method == 'pivotal':
        low = 2 * stat_val - _np.percentile(bootstrap_dist, 100 * (1 - alpha / 2.))
        val = stat_val
        high = 2 * stat_val - _np.percentile(bootstrap_dist, 100 * (alpha / 2.))
    elif ci_method == 'perc':
        # check if All values of bootstrap_dist are equal and if YES -
        # display a warning and do not calculate CIs - like boot.ci in R
        if _all_the_same(bootstrap_dist):
            print(f'All values of t are equal to {bootstrap_dist[0]}. Cannot calculate confidence intervals')
            low = None
            high = None
        else:
            bd = bootstrap_dist[bootstrap_dist != _np.array([None])]
            low = _np.percentile(bd, 100 * (alpha / 2.), interpolation='linear')
            high = _np.percentile(bd, 100 * (1 - alpha / 2.), interpolation='linear')
        val = stat_val
    else:
        low = None
        val = None
        high = None
    return BootstrapResults(lower_bound=low,
                            value=val,
                            upper_bound=high)


def _get_confidence_interval_and_value_eclv(bootstrap_dist, stat_val, alpha, ci_method):
    """Get the bootstrap confidence interval for a given distribution for the Economic Cost Loss Relative Value
        Args:
            bootstrap_dist: numpy array of bootstrap results from
                bootstrap_distribution() or  bootstrap_distribution_cbb()
                or bootstrap_ab_distribution()
            stat_val: The overall statistic that this method is attempting to
                calculate error bars for.
            alpha: The alpha value for the confidence intervals.
            ci_method: if true, use the pivotal method. if false, use the
                percentile method.
    """

    # TODO Only percentile method for the confident intervals is implemented

    if stat_val is None:
        val = None
        low = None
        high = None

    else:
        bd = bootstrap_dist[bootstrap_dist != _np.array([None])]
        all_values = []
        for dist_member in bd:
            all_values.append(dist_member['V'].tolist())

        all_values_np = _np.array(all_values)
        steps_len = len(stat_val['cl'])
        none_in_values = len(stat_val['V']) != sum(x is not None for x in stat_val['V'])
        stat_btcl = [None] * steps_len
        stat_btcu = [None] * steps_len

        for ind in range(steps_len):
            low = None
            high = None
            column = all_values_np[:, ind]
            if ci_method == 'pivotal':
                low = 2 * stat_val - _np.percentile(column, 100 * (1 - alpha / 2.))
                high = 2 * stat_val - _np.percentile(column, 100 * (alpha / 2.))

            elif ci_method == 'perc':
                if _all_the_same(column):
                    print(f'All values of t are equal to {column[0]}. Cannot calculate confidence intervals')
                    low = None
                    high = None
                else:
                    if none_in_values:
                        low = _np.percentile(column, 100 * (alpha / 2.), interpolation='linear')
                        high = _np.percentile(column, 100 * (1 - alpha / 2.), interpolation='linear')
            stat_btcl[ind] = low
            stat_btcu[ind] = high

        val = stat_val
        low = stat_btcl
        high = stat_btcu

    return BootstrapResults(lower_bound=low,
                            value=val,
                            upper_bound=high)


def flatten(lis):
    for item in lis:
        if isinstance(item, Iterable) and not isinstance(item, float):
            for x in flatten(item):
                yield x
        else:
            yield item


def _bootstrap_sim_cbb(values_lists, stat_func_lists, num_iterations,
                       iteration_batch_size, seed, block_length=1):
    """Returns simulated bootstrap distribution. Can do the independent and identically distributed (IID)
        or Circular Block Bootstrap (CBB) methods depending on the block_length
        Args:
            values_lists: numpy array  of values to bootstrap

            stat_func_lists: statistic to bootstrap

            num_iterations: number of bootstrap iterations to run. The higher this
            number the more sure you can be about the stability your bootstrap.

            iteration_batch_size: The bootstrap sample can generate very large
            matrices. This argument limits the memory footprint by
            batching bootstrap rounds. If unspecified the underlying code
            will produce a matrix of len(values) x num_iterations. If specified
            the code will produce sets of len(values) x iteration_batch_size
            (one at a time) until num_iterations have been simulated.

            seed: random seed

            block_length: number giving the desired block lengths.
                Default (block.length = 1) is to do IID resamples.
                Should be longer than the length of dependence in the data,
                but much shorter than the size of the data.
    """

    if seed is not None:
        _np.random.seed(seed)

    num_iterations = int(num_iterations)
    iteration_batch_size = int(iteration_batch_size)

    results = [[] for _ in values_lists]

    for rng in range(0, num_iterations, iteration_batch_size):
        max_rng = min(iteration_batch_size, num_iterations - rng)

        values_sims = _generate_distributions_cbb(values_lists, max_rng, block_length)

        for i, values_sim, stat_func in zip(range(len(values_sims)), values_sims, stat_func_lists):
            results[i].extend(stat_func(values_sim))

    return _np.array(results)


def _generate_distributions_cbb(values_lists, num_iterations, block_length=1):
    values_shape = values_lists[0].shape[0]
    ids = _np.random.choice(
        values_shape,
        (num_iterations, values_shape),
        replace=True
    )

    def apply_cbb(row):
        """
        Applyes Circular Block Bootstrap (CBB) method to each row
        :param row:
        """
        counter = 0
        init_val = row[0]
        for ind, val in enumerate(row):
            if counter == 0:
                # save a 1st value for the block
                init_val = val
            else:
                # calculate current value by adding the counter to the initial value
                new_val = init_val + counter
                # the value should not be bigger then the size of the row
                if new_val > len(row) - 1:
                    new_val = new_val - len(row)
                row[ind] = new_val
            counter = counter + 1
            if counter == block_length:
                # start a new block
                counter = 0
        return row

    if block_length > 1:
        # uss CBB
        ids = _np.apply_along_axis(apply_cbb, axis=1, arr=ids)

    results = [values[ids] for values in values_lists]
    return results


def _all_the_same(elements):
    """Checks if all elements in the numpy array are the same

         Args:
            elements: numpy array

        Returns:
            True if array is empty or all elements are the same
            False if the array has different elements
    """
    if elements.size == 0:
        return True
    try:
        result = len(_np.unique(elements[elements != _np.array([None])])) == 1
    except TypeError:
        result = False
    return result


def _validate_arrays(values_lists):
    t = values_lists[0]
    t_type = type(t)
    if not isinstance(t, _sparse.csr_matrix) and not isinstance(t, _np.ndarray):
        raise ValueError(('The arrays must either be of type '
                          'scipy.sparse.csr_matrix or numpy.array'))

    for _, values in enumerate(values_lists[1:]):
        if not isinstance(values, t_type):
            raise ValueError('The arrays must all be of the same type')

        if t.shape != values.shape:
            raise ValueError('The arrays must all be of the same shape')

        if isinstance(t, _sparse.csr_matrix):
            if values.shape[0] > 1:
                raise ValueError(('The sparse matrix must have shape 1 row X N'
                                  ' columns'))

    if isinstance(t, _sparse.csr_matrix):
        if _needs_sparse_unification(values_lists):
            raise ValueError(('The non-zero entries in the sparse arrays'
                              ' must be aligned'))


def _needs_sparse_unification(values_lists):
    non_zeros = values_lists[0] != 0

    for v in values_lists:
        v_nz = v != 0
        non_zeros = (non_zeros + v_nz) > 0

    non_zero_size = non_zeros.sum()

    for v in values_lists:
        if non_zero_size != v.data.shape[0]:
            return True

    return False


def mean(values, axis=1):
    """Returns the mean of each row of a matrix"""
    if isinstance(values, _sparse.csr_matrix):
        ret = values.mean(axis=axis)
        return ret.A1
    else:
        return _np.mean(_np.asmatrix(values), axis=axis).A1


def sum(values, axis=1):
    """Returns the sum of each row of a matrix"""
    if isinstance(values, _sparse.csr_matrix):
        ret = values.sum(axis=axis)
        return ret.A1
    else:
        return _np.sum(_np.asmatrix(values), axis=axis).A1


def median(values, axis=1):
    """Returns the sum of each row of a matrix"""
    if isinstance(values, _sparse.csr_matrix):
        ret = values.median(axis=axis)
        return ret.A1
    else:
        return _np.median(_np.asmatrix(values), axis=axis).A1


def std(values, axis=1):
    """ Returns the std of each row of a matrix"""
    if isinstance(values, _sparse.csr_matrix):
        ret = values.std(axis=axis)
        return ret.A1
    else:
        return _np.std(_np.asmatrix(values), axis=axis).A1



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: agg_stat_event_equalize.py

How to use:
 - Call from other Python function
        AGG_STAT_EVENT_EQUALIZE = AggStatEventEqualize(PARAMS)
        AGG_STAT_EVENT_EQUALIZE.calculate_values()
        where PARAMS – a dictionary with data description parameters including
        location of input and output data.
        The structure is similar to Rscript template

 - Run as a stand-alone script
        python agg_stat_event_equalize.py <parameters_file>
        where - <parameters_file> is YAML file with parameters
        and environment variable should be set to PYTHONPATH=<path_to_METcalcpy>

 - Run from Java
        proc = Runtime.getRuntime().exec(
                “python agg_stat_event_equalize.eqz.py <parameters_file>”,
                new String[]{”PYTHONPATH=<path_to_METcalcpy>”},
                new File(System.getProperty("user.home")));

"""
import argparse
import sys
import itertools
import pandas as pd
import yaml
import numpy as np

from metcalcpy import GROUP_SEPARATOR
from metcalcpy.event_equalize import event_equalize

class AggStatEventEqualize:
    """A class that performs event equalisation logic on input data
           with MODE and MTD attribute statistics

           All parameters including data description and location is in the parameters dictionary
        Usage:
            initialise this call with the parameters dictionary and then
            calls perform_ee method
            This method will execute EE and save the result to the file
                AGG_STAT_EVENT_EQUALIZE = AggStatEventEqualize(PARAMS)
                AGG_STAT_EVENT_EQUALIZE.calculate_values()
           """

    def __init__(self, in_params):
        self.params = in_params

        self.input_data = pd.read_csv(
            self.params['agg_stat_input'],
            header=[0],
            sep='\t'
        )

        self.column_names = self.input_data.columns.values
        self.series_data = None

    def calculate_values(self):
        """Performs event equalisation if needed and saves equalized data to the file.
        """
        if not self.input_data.empty:

            # list all fixed variables
            if 'fixed_vars_vals_input' in self.params:
                fix_vals_permuted_list = []
                for key in self.params['fixed_vars_vals_input']:
                    vals_permuted = list(itertools.product(*self.params['fixed_vars_vals_input'][key].values()))
                    fix_vals_permuted_list.append(vals_permuted)

                fix_vals_permuted = [item for sublist in fix_vals_permuted_list for item in sublist]

            else:
                fix_vals_permuted = []

            # perform EE for each forecast variable on y1 axis
            output_ee_data = self.run_ee_on_axis(fix_vals_permuted, '1')

            # if the second Y axis is present - run event equalizer on Y1
            # and then run event equalizer on Y1 and Y2 equalized data
            if self.params['series_val_2']:
                output_ee_data_2 = self.run_ee_on_axis(fix_vals_permuted, '2')

                output_ee_data = output_ee_data.drop('equalize', axis=1)
                output_ee_data_2 = output_ee_data_2.drop('equalize', axis=1)
                all_ee_records = output_ee_data.append(output_ee_data_2).reindex()
                all_series_vars = {}
                for key in self.params['series_val_2']:
                    all_series_vars[key] = np.unique(self.params['series_val_2'][key]
                                                     + self.params['series_val_2'][key])

                output_ee_data = event_equalize(all_ee_records, self.params['indy_var'],
                                                all_series_vars,
                                                list(self.params['fixed_vars_vals_input'].keys()),
                                                fix_vals_permuted, True,
                                                False)

        else:
            output_ee_data = pd.DataFrame()

        header = True
        mode = 'w'
        output_ee_data.to_csv(self.params['agg_stat_output'],
                                           index=None, header=header, mode=mode,
                                           sep="\t", na_rep="NA")

    def run_ee_on_axis(self, fix_vals_permuted, axis='1'):
        """Performs event equalisation against previously calculated cases for the selected axis
           Returns:
               A data frame that contains equalized records
       """

        output_ee_data = pd.DataFrame()
        for series_var, series_var_vals in self.params['series_val_' + axis].items():
            # ungroup series value
            series_var_vals_no_group = []
            for val in series_var_vals:
                split_val = val.split(GROUP_SEPARATOR)
                series_var_vals_no_group.extend(split_val)

            # filter input data based on fcst_var, statistic and all series variables values
            series_data_for_ee = self.input_data[
                self.input_data[series_var].isin(series_var_vals_no_group)
            ]
            # perform EE on filtered data
            series_data_after_ee = \
                event_equalize(series_data_for_ee, self.params['indy_var'],
                               self.params['series_val_' + axis],
                               list(self.params['fixed_vars_vals_input'].keys()),
                               fix_vals_permuted, True, False)

            # append EE data to result
            if output_ee_data.empty:
                output_ee_data = series_data_after_ee
            else:
                output_ee_data = output_ee_data.append(series_data_after_ee)
        return output_ee_data


if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(description='List of agg_stat_event_equalize arguments')
    PARSER.add_argument("parameters_file", help="Path to YAML parameters file",
                        type=argparse.FileType('r'),
                        default=sys.stdin)
    ARGS = PARSER.parse_args()
    PARAMS = yaml.load(ARGS.parameters_file, Loader=yaml.FullLoader)

    AGG_STAT_EVENT_EQUALIZE = AggStatEventEqualize(PARAMS)
    AGG_STAT_EVENT_EQUALIZE.calculate_values()
